import os
from typing import Literal
from ..BaseLoader import BaseLoader

import numpy as np


MnistType = Literal["mnist", "fmnist", "kmnist"]

label_dict = {
    "mnist": {
        0: "0",
        1: "1",
        2: "2",
        3: "3",
        4: "4",
        5: "5",
        6: "6",
        7: "7",
        8: "8",
        9: "9",
    },
    "fmnist": {
        0: "T-shirt/top",
        1: "Trouser",
        2: "Pullover",
        3: "Dress",
        4: "Coat",
        5: "Sandal",
        6: "Shirt",
        7: "Sneaker",
        8: "Bag",
        9: "Ankle boot",
    },
    "kmnist": {
        0: "O",
        1: "Ki",
        2: "Su",
        3: "TSU",
        4: "Na",
        5: "Ha",
        6: "Ma",
        7: "Ya",
        8: "Re",
        9: "Wo",
    },
}


class MnistSeriesLoader(BaseLoader):
    def __init__(self, name: MnistType = "mnist"):
        super().__init__()
        self.name = name
        self.base_path = os.path.join(
            os.path.dirname(os.path.abspath(__file__)), self.name
        )
        self.load_data()

    def load_raw_data(self):
        self._data = np.load(os.path.join(self.base_path, "data.npy"))
        self._label = np.load(os.path.join(self.base_path, "label.npy"))
        self._data = self.scale_data(self._data)
        self._legend = [label_dict[self.name][i] for i in range(10)]

        self._precomputed_knn = self.get_precomputed_knn(self._data)

        self.save_data(self.base_path)



import json
import os
from typing import Literal
from ..BaseLoader import BaseLoader

import numpy as np


TextEmbtypes = Literal["20ng", "ag_news", "amazon_polarity", "yelp_review"]


class TextEmbeddingLoader(BaseLoader):
    def __init__(self, name: TextEmbtypes = "20ng"):
        super().__init__()
        self.name = name
        self.base_path = os.path.join(
            os.path.dirname(os.path.abspath(__file__)), self.name
        )
        self.load_data()

    def load_raw_data(self):
        self._data = np.load(os.path.join(self.base_path, "data.npy"))
        self._label = np.load(os.path.join(self.base_path, "label.npy"))
        self._legend = json.load(open(os.path.join(self.base_path, "legend.json"))).get(
            "legend"
        )

        self._precomputed_knn = self.get_precomputed_knn(self._data)

        self.save_data(self.base_path)



from dataclasses import dataclass
from typing import TypedDict

import numpy as np


class TPoint(TypedDict):
    x: float
    y: float
    r: float


@dataclass
class GhostPointModel:
    id: int
    points: list[TPoint]
    label: str | None

    def to_dict(self):
        return {
            "id": self.id,
            "coords": self.points,
            "label": self.label,
        }


class GhostEmbedding:
    @staticmethod
    def build_model(
        ghost_embedding: np.ndarray,  # shape (n_points, n_ghosts, 2)
        init_radii: np.ndarray,  # shape (n_points, n_ghosts)
        label: np.ndarray | list[str] | None,  # shape (n_points,)
    ) -> list[dict]:
        if label is None:
            label = [None] * len(ghost_embedding)

        return [
            GhostPointModel(
                id=i,
                points=[{"x": x, "y": y, "r": r} for (x, y), r in zip(g, r)],
                label=str(l),
            ).to_dict()
            for i, (g, r, l) in enumerate(zip(ghost_embedding, init_radii, label))
        ]

        # return [
        #     GhostPointModel(
        #         id=i, coords=[{"x": x, "y": y} for x, y in g], label=str(l)
        #     ).to_dict()
        #     for i, (g, l) in enumerate(zip(ghost_embedding, label))
        # ]



from dataclasses import dataclass, field, asdict, replace
from typing import Optional

import numpy as np


@dataclass(frozen=True)
class Result:
    opt_time: float = field(default=0.0)
    distance_list: np.ndarray = field(default_factory=lambda: np.array([]))
    threshold_list: np.ndarray = field(default_factory=lambda: np.array([]))
    init_radii: np.ndarray = field(default_factory=lambda: np.array([]))


# Lazy initialization of config
_result: Optional[Result] = Result()


def get_results() -> Result:
    global _result
    if _result is None:
        raise ValueError(
            "Result not set. Please set the results using `set_results` or `set_eval_results`."
        )
    return _result


def set_results(
    opt_time=None,
    distance_list=None,
    threshold_list=None,
    init_radii=None,
    reinit: bool = False,
) -> None:
    global _result

    if reinit:
        _result = Result()
        return

    _result = replace(
        _result,
        opt_time=opt_time if opt_time is not None else _result.opt_time,
        distance_list=(
            distance_list if distance_list is not None else _result.distance_list
        ),
        threshold_list=(
            threshold_list if threshold_list is not None else _result.threshold_list
        ),
        init_radii=(init_radii if init_radii is not None else _result.init_radii),
    )


__all__ = [
    "get_results",
    "set_results",
]



from typing import Literal, TypedDict
import numpy as np


class THyperparameter(TypedDict):
    n_ghosts: int
    radii: float
    sensitivity: float
    ghost_gen: float
    init_dropping: float
    mov_avg_weight: float


class TResult(TypedDict):
    opt_time: float
    original_embedding: np.ndarray
    ghost_embedding: np.ndarray
    ghost_mask: np.ndarray
    unstable_ghosts: np.ndarray
    distance_list: np.ndarray
    threshold_list: np.ndarray


Tbenchmark = Literal[
    "accuracy_dropping",
    "accuracy_SH",
    "time_with_dropping",
    "time_with_SH",
    "time_original_GU",
]


__all__ = [
    "THyperparameter",
    "TResult",
    "Tbenchmark",
]



from typing import Literal, TypedDict

import numpy as np


TprecomputedKnn = (
    tuple[np.ndarray, np.ndarray]
    | tuple[np.ndarray, np.ndarray, np.ndarray]
    | tuple[None, None]
    | tuple[None, None, None]
)

TDataName = Literal[
    "mnist",
    "fmnist",
    "kmnist",
    "celegans",
    "ionosphere",
    "optical_recognition",
    "raisin",
    "htru2",
    "parishousing",
    "cnae9",
    "20ng",
    "ag_news",
    "amazon_polarity",
    "yelp_review",
]


class DataModel(TypedDict):
    data: np.ndarray
    label: np.ndarray | None
    legend: dict | None
    precomputed_knn: TprecomputedKnn = (None, None)



import os

from ..BaseLoader import BaseLoader

import numpy as np
import pandas as pd


class CelegansLoader(BaseLoader):
    def __init__(self):
        super().__init__()
        self.name = "celegans"
        self.base_path = os.path.dirname(os.path.abspath(__file__))
        self.load_data()

    def load_raw_data(self):
        self._data = np.load(os.path.join(self.base_path, "data.npy"))
        self._label = np.load(os.path.join(self.base_path, "label.npy"))
        self._data = self.scale_data(self._data)
        self._legend = [
            "ASE, ASJ, AUA",
            "ASH",
            "AWC",
            "ADL",
            "ASG, AWA",
            "ASK",
            "ASI",
            "AFD",
            "ADF, AWB",
            "Not annotated",
        ]

        self._precomputed_knn = self.get_precomputed_knn(self._data)

        self.save_data(self.base_path)

    def _load_raw_data(self):
        preprocessed_path = os.path.join(self.base_path, "celegans_proccessed.csv")
        metadata_path = os.path.join(self.base_path, "celegans_metadata.csv")

        df = pd.read_csv(preprocessed_path, index_col=0)
        df2 = pd.read_csv(metadata_path, index_col=0)
        cell_types = df2["cell_type"]
        cell_types = cell_types.fillna("")

        ctype_to_lineage = {
            "": "Not annotated",
            ##
            "Neuroblast_ASE_ASJ_AUA": "ASE, ASJ, AUA",
            "Neuroblast_ASJ_AUA": "ASE, ASJ, AUA",
            "ASE_parent": "ASE, ASJ, AUA",
            "ASE": "ASE, ASJ, AUA",
            "ASEL": "ASE, ASJ, AUA",
            "ASER": "ASE, ASJ, AUA",
            "ASJ": "ASE, ASJ, AUA",
            "AUA": "ASE, ASJ, AUA",
            ##
            "Neuroblast_ASG_AWA": "ASG, AWA",
            "ASG_AWA": "ASG, AWA",
            "ASG": "ASG, AWA",
            "AWA": "ASG, AWA",
            ##
            "Neuroblast_ADF_AWB": "ADF, AWB",
            "ADF_AWB": "ADF, AWB",
            "ADF": "ADF, AWB",
            "AWB": "ADF, AWB",
            ##
            "AWC": "AWC",
            "AWC_ON": "AWC",
            ##
            "Neuroblast_AFD_RMD": "AFD",
            "AFD": "AFD",
            ##
            "ADL_parent": "ADL",
            "ADL": "ADL",
            ##
            "ASH": "ASH",
            ##
            "ASI_parent": "ASI",
            "ASI": "ASI",
            ##
            "ASK_parent": "ASK",
            "ASK": "ASK",
        }

        self._data = self.scale_data(np.array(df))
        self._label = np.array([ctype_to_lineage[celltype] for celltype in cell_types])
        self._legend = [
            "ASE, ASJ, AUA",
            "ASH",
            "AWC",
            "ADL",
            "ASG, AWA",
            "ASK",
            "ASI",
            "AFD",
            "ADF, AWB",
            "Not annotated",
        ]
        self._precomputed_knn = self.get_precomputed_knn(self._data)

        self.save_data(self.base_path)



from datetime import datetime
import logging

from benchmark.runner import benchmark_v0, benchmark_v1, benchmark_v2
from benchmark.hyperparameters import generate_hyperparameter_comb
from benchmark.utils import run_GhostUMAP, measure_accuracy
from benchmark.save_manager import save_embeddings, save_results
from rdumap.data import DataLoader


def main(
    data_name: str,
    base_settings: dict,
    param_grid: dict,
    iterations: int = 3,
    version="v1",
    results_dir="results",
):
    # Load dataset
    print(version)

    logging.info(f"Loading dataset: {data_name}")
    dl = DataLoader(data_name)
    X, y, legend, precomputed_knn = dl.get_data().values()

    logging.info("Generating hyperparameter combinations.")
    hpram_comb = generate_hyperparameter_comb(base_settings, param_grid)

    benchmark_func = {
        "v0": benchmark_v0,
        "v1": benchmark_v1,
        "v2": benchmark_v2,
    }.get(version, benchmark_v2)

    for hprams in hpram_comb:
        if hprams["ghost_gen"] >= hprams["init_dropping"]:
            logging.info(
                f"Skipping combination due to {hprams['ghost_gen']} >= {hprams['init_dropping']}"
            )
            continue

        logging.info(f"Running benchmark for parameter set: {hprams}")

        results = benchmark_func(
            data_name,
            X,
            precomputed_knn,
            hprams,
            iterations,
        )
        save_results(data_name, results, results_dir=results_dir)
        logging.info("Results saved for current parameter set.")


if __name__ == "__main__":
    logging.basicConfig(
        filename="benchmark.log",
        format="%(levelname)s: %(message)s",
        level=logging.INFO,
        filemode="a",
    )

    timestamp = datetime.now().strftime("%Y%m%d_%H%M%S")

    data_name = [
        # "ionosphere",
        # "raisin",
        "celegans",
        "parishousing",
        "htru2",
        "optical_recognition",
        "mnist",
        "fmnist",
        "kmnist",
        "cnae9",
        "20ng",
        "ag_news",
        "amazon_polarity",
        # "yelp_review",
    ]
    # data_name = ["raisin"]

    base_settings = {
        "radii": 0.1,
        "sensitivity": 0.9,
        "mov_avg_weight": 0.9,
        "ghost_gen": 0.2,
        "init_dropping": 0.4,
    }
    param_grid = {
        "n_ghosts": [8, 16, 32],
        # "n_ghosts": [8],
    }

    for data in data_name:
        logging.info(f"Starting benchmark for dataset: {data}")
        main(
            data,
            base_settings,
            param_grid,
            iterations=1,
            version="v2",
            results_dir=f"results_{timestamp}",
        )
        logging.info(f"Completed benchmark for dataset: {data}")



import os
import json
from typing import Literal

from ..BaseLoader import BaseLoader

import numpy as np
import pandas as pd

from umap import UMAP

TUciml = Literal["ionosphere", "optical_recognition", "raisin", "htru2"]

data_id = {
    "ionosphere": 52,
    "optical_recognition": 80,
    "raisin": 850,
    "htru2": 372,
}

legend_to_label = {
    "ionosphere": {
        "g": 0,
        "b": 1,
    },
    "optical_recognition": {
        "0": 0,
        "1": 1,
        "2": 2,
        "3": 3,
        "4": 4,
        "5": 5,
        "6": 6,
        "7": 7,
        "8": 8,
        "9": 9,
    },
    "raisin": {
        "Kecimen": 0,
        "Besni": 1,
    },
    "htru2": {
        "0": 0,
        "1": 1,
    },
}


class UcimlLoader(BaseLoader):
    def __init__(self, name: TUciml = "ionosphere"):
        super().__init__()
        self.name = name
        self.base_path = os.path.join(
            os.path.dirname(os.path.abspath(__file__)), self.name
        )
        self.load_data()

    def load_raw_data(self):
        from ucimlrepo import fetch_ucirepo

        fetched_data = fetch_ucirepo(id=data_id[self.name])

        # data (as pandas dataframes)
        X = fetched_data.data.features.values
        y = fetched_data.data.targets.values.flatten()

        self._data = self.scale_data(X)
        self._label = np.array([legend_to_label[self.name][str(label)] for label in y])
        self._legend = list(legend_to_label[self.name].keys())

        self._precomputed_knn = self.get_precomputed_knn(self._data)

        if not os.path.exists(self.base_path):
            os.makedirs(self.base_path)

        self.save_data(self.base_path)



from dataclasses import asdict
import numpy as np

from .models import THyperparameter, TResult, Tbenchmark
from rdumap.ghostumap import GhostUMAP


def run_GhostUMAP(
    X: np.ndarray,
    hyperparameters: THyperparameter,
    bm_type: Tbenchmark,
    precomputed_knn=(None, None, None),
    distance=0.1,
) -> TResult:
    """
    Run grid search.
    """

    gu = GhostUMAP(precomputed_knn=precomputed_knn)

    O, G, alive_ghosts = gu.fit_transform(X, bm_type=bm_type, **hyperparameters)
    unstable_ghosts = (
        gu.get_unstable_ghosts(
            distance=distance,
            sensitivity=hyperparameters["sensitivity"],
        )
        if "accuracy" in bm_type
        else np.array([])
    )
    result = asdict(gu.get_results())

    result["original_embedding"] = O
    result["ghost_embedding"] = G
    result["alive_ghosts"] = alive_ghosts
    result["unstable_ghosts"] = unstable_ghosts

    return result


def measure_accuracy(y_true, y_pred):
    TP = np.sum(np.logical_and(y_true == 1, y_pred == 1))
    FP = np.sum(np.logical_and(y_true == 0, y_pred == 1))
    FN = np.sum(np.logical_and(y_true == 1, y_pred == 0))
    TN = np.sum(np.logical_and(y_true == 0, y_pred == 0))

    precision = TP / (TP + FP) if TP + FP > 0 else 0
    recall = TP / (TP + FN) if TP + FN > 0 else 0

    f1 = (
        2 * (precision * recall) / (precision + recall) if precision + recall > 0 else 0
    )

    return f1, precision, recall


__all__ = ["run_GhostUMAP", "measure_accuracy"]



import os
import json

from ..BaseLoader import BaseLoader

import numpy as np

from umap import UMAP


class ParisHousingLoader(BaseLoader):
    def __init__(self):
        super().__init__()
        self.name = "parishousing"
        self.base_path = os.path.dirname(os.path.abspath(__file__))
        self.load_data()

    def load_raw_data(self):
        import pandas as pd

        data = pd.read_csv(os.path.join(self.base_path, "ParisHousingClass.csv"))

        X = data.drop("category", axis=1).values
        y = data["category"].values

        self._data = self.scale_data(X)
        self._label = np.array([0 if i == "Basic" else 1 for i in y])
        self._legend = ["Basic", "Luxury"]
        self._precomputed_knn = self.get_precomputed_knn(self._data)

        self.save_data(self.base_path)



import itertools


def generate_hyperparameter_comb(
    base_settings: dict,
    param_grid: dict,
):
    """
    Generate hyperparameter combinations for grid search.
    """
    param_combinations = list(itertools.product(*param_grid.values()))
    keys = param_grid.keys()

    combined_settings = []
    for values in param_combinations:
        settings = base_settings.copy()
        settings.update(dict(zip(keys, values)))
        combined_settings.append(settings)

    return combined_settings


def make_dir_name(hyperparameters):
    hpram_keys = [
        "n_ghosts",
        "radii",
        "sensitivity",
        "ghost_gen",
        "init_dropping",
        "mov_avg_weight",
    ]
    hpram_abbr = ["ng", "r", "sens", "gg", "init", "mov"]

    dir_name = "_".join(
        [
            f"{abbr}_{val}"
            for abbr, val in zip(
                hpram_abbr, [hyperparameters[key] for key in hpram_keys]
            )
        ]
    )

    return dir_name


__all__ = ["generate_hyperparameter_comb", "make_dir_name"]



from dataclasses import dataclass
from typing import List, Dict, Any
import numpy as np
from .OriginalEmbedding import OriginalEmbedding
from .GhostEmbedding import GhostEmbedding


class EmbeddingSet:
    def __init__(
        self,
        original_embedding: np.ndarray,
        ghost_embedding: np.ndarray,
        neighbors: np.ndarray,
        r: float,
        init_radii: np.ndarray,
        label: np.ndarray | list | None,
        title: str,
        legend: List[str] | None,
        colors: Dict[str, str] | None = None,
    ):
        self.original_embedding = original_embedding
        self.ghost_embedding = ghost_embedding
        self.neighbors = neighbors
        self.r = r
        self.init_radii = init_radii

        self.label = label
        self.title = title
        self.legend = legend
        self.colors = colors
        self.n_ghosts = ghost_embedding.shape[1]

    def to_dict(self) -> Dict[str, Any]:
        return {
            "original_embedding": OriginalEmbedding.build_model(
                self.original_embedding,
                self.ghost_embedding,
                self.neighbors,
                self.label,
            ),
            "ghost_embedding": GhostEmbedding.build_model(
                self.ghost_embedding, self.init_radii, self.label
            ),
            "r": self.r,
            "title": self.title,
            "legend": self.legend or self.generate_default_legend(self.label),
            "colors": self.colors or {},
            "n_ghosts": self.n_ghosts,
        }

    @staticmethod
    def generate_default_legend(label: np.ndarray | list | None) -> List[str]:
        if label is None:
            return []

        return [str(l) for l in sorted(set(label))]



import time
from warnings import warn
import joblib


from sklearn.decomposition import PCA, TruncatedSVD
from sklearn.metrics import pairwise_distances
from sklearn.neighbors import KDTree
from sklearn.utils import check_array, check_random_state

from umap import UMAP

from umap import distances as dist
from umap import sparse as sparse
from umap.umap_ import (
    find_ab_params,
    noisy_scale_coords,
    make_epochs_per_sample,
    INT32_MIN,
    INT32_MAX,
    reset_local_connectivity,
    raise_disconnected_warning,
    general_simplicial_set_intersection,
    fuzzy_simplicial_set,
    discrete_metric_simplicial_set_intersection,
    nearest_neighbors,
)
from umap.utils import (
    ts,
    csr_unique,
)
from umap.spectral import spectral_layout, tswspectral_layout


import numpy as np
import numba
import scipy.sparse
from scipy.sparse import tril as sparse_tril, triu as sparse_triu


from pynndescent.distances import named_distances as pynn_named_distances
from pynndescent.sparse import sparse_named_distances as pynn_sparse_named_distances

from .utils import get_distance, compute_distances, drop_ghosts

from .layouts import optimize_layout_euclidean

from .configs import get_config as _get_config, set_config
from .results import get_results as _get_results, set_results
from .widget import Widget
from IPython.display import display


def simplicial_set_embedding(
    data,
    n_ghosts,
    graph,
    n_components,
    initial_alpha,
    a,
    b,
    gamma,
    negative_sample_rate,
    n_epochs,
    init,
    random_state,
    metric,
    metric_kwds,
    parallel=False,
    verbose=False,
    tqdm_kwds=None,
):
    """Perform a fuzzy simplicial set embedding, using a specified
    initialisation method and then minimizing the fuzzy set cross entropy
    between the 1-skeletons of the high and low dimensional fuzzy simplicial
    sets.

    Parameters
    ----------
    data: array of shape (n_samples, n_features)
        The source data to be embedded by UMAP.

    graph: sparse matrix
        The 1-skeleton of the high dimensional fuzzy simplicial set as
        represented by a graph for which we require a sparse matrix for the
        (weighted) adjacency matrix.

    n_components: int
        The dimensionality of the euclidean space into which to embed the data.

    initial_alpha: float
        Initial learning rate for the SGD.

    a: float
        Parameter of differentiable approximation of right adjoint functor

    b: float
        Parameter of differentiable approximation of right adjoint functor

    gamma: float
        Weight to apply to negative samples.

    negative_sample_rate: int (optional, default 5)
        The number of negative samples to select per positive sample
        in the optimization process. Increasing this value will result
        in greater repulsive force being applied, greater optimization
        cost, but slightly more accuracy.

    n_epochs: int (optional, default 0), or list of int
        The number of training epochs to be used in optimizing the
        low dimensional embedding. Larger values result in more accurate
        embeddings. If 0 is specified a value will be selected based on
        the size of the input dataset (200 for large datasets, 500 for small).
        If a list of int is specified, then the intermediate embeddings at the
        different epochs specified in that list are returned in
        ``aux_data["embedding_list"]``.

    init: string
        How to initialize the low dimensional embedding. Options are:

            * 'spectral': use a spectral embedding of the fuzzy 1-skeleton
            * 'random': assign initial embedding positions at random.
            * 'pca': use the first n_components from PCA applied to the input data.
            * A numpy array of initial embedding positions.

    random_state: numpy RandomState or equivalent
        A state capable being used as a numpy random state.

    metric: string or callable
        The metric used to measure distance in high dimensional space; used if
        multiple connected components need to be layed out.

    metric_kwds: dict
        Key word arguments to be passed to the metric function; used if
        multiple connected components need to be layed out.

    densmap: bool
        Whether to use the density-augmented objective function to optimize
        the embedding according to the densMAP algorithm.

    densmap_kwds: dict
        Key word arguments to be used by the densMAP optimization.

    output_dens: bool
        Whether to output local radii in the original data and the embedding.

    output_metric: function
        Function returning the distance between two points in embedding space and
        the gradient of the distance wrt the first argument.

    output_metric_kwds: dict
        Key word arguments to be passed to the output_metric function.

    euclidean_output: bool
        Whether to use the faster code specialised for euclidean output metrics

    parallel: bool (optional, default False)
        Whether to run the computation using numba parallel.
        Running in parallel is non-deterministic, and is not used
        if a random seed has been set, to ensure reproducibility.

    verbose: bool (optional, default False)
        Whether to report information on the current progress of the algorithm.

    tqdm_kwds: dict
        Key word arguments to be used by the tqdm progress bar.

    Returns
    -------
    embedding: array of shape (n_samples, n_components)
        The optimized of ``graph`` into an ``n_components`` dimensional
        euclidean space.

    aux_data: dict
        Auxiliary output returned with the embedding. When densMAP extension
        is turned on, this dictionary includes local radii in the original
        data (``rad_orig``) and in the embedding (``rad_emb``).
    """
    graph = graph.tocoo()
    graph.sum_duplicates()
    n_vertices = graph.shape[1]

    # For smaller datasets we can use more epochs
    if graph.shape[0] <= 10000:
        default_epochs = 500
    else:
        default_epochs = 200

    if n_epochs is None:
        n_epochs = default_epochs

    # If n_epoch is a list, get the maximum epoch to reach
    n_epochs_max = max(n_epochs) if isinstance(n_epochs, list) else n_epochs

    if n_epochs_max > 10:
        graph.data[graph.data < (graph.data.max() / float(n_epochs_max))] = 0.0
    else:
        graph.data[graph.data < (graph.data.max() / float(default_epochs))] = 0.0

    graph.eliminate_zeros()

    if isinstance(init, str) and init == "random":
        embedding = random_state.uniform(
            low=-10.0, high=10.0, size=(graph.shape[0], n_components)
        ).astype(np.float32)
    elif isinstance(init, str) and init == "pca":
        if scipy.sparse.issparse(data):
            pca = TruncatedSVD(n_components=n_components, random_state=random_state)
        else:
            pca = PCA(n_components=n_components, random_state=random_state)
        embedding = pca.fit_transform(data).astype(np.float32)
        embedding = noisy_scale_coords(
            embedding, random_state, max_coord=10, noise=0.0001
        )
    elif isinstance(init, str) and init == "spectral":
        embedding = spectral_layout(
            data,
            graph,
            n_components,
            random_state,
            metric=metric,
            metric_kwds=metric_kwds,
        )
        # We add a little noise to avoid local minima for optimization to come
        embedding = noisy_scale_coords(
            embedding, random_state, max_coord=10, noise=0.0001
        )

    elif isinstance(init, str) and init == "tswspectral":
        embedding = tswspectral_layout(
            data,
            graph,
            n_components,
            random_state,
            metric=metric,
            metric_kwds=metric_kwds,
        )
        embedding = noisy_scale_coords(
            embedding, random_state, max_coord=10, noise=0.0001
        )
    else:
        init_data = np.array(init)
        if len(init_data.shape) == 2:
            if np.unique(init_data, axis=0).shape[0] < init_data.shape[0]:
                tree = KDTree(init_data)
                dist, ind = tree.query(init_data, k=2)
                nndist = np.mean(dist[:, 1])
                embedding = init_data + random_state.normal(
                    scale=0.001 * nndist, size=init_data.shape
                ).astype(np.float32)
            else:
                embedding = init_data

    epochs_per_sample = make_epochs_per_sample(graph.data, n_epochs_max)

    head = graph.row
    tail = graph.col
    weight = graph.data

    rng_state = random_state.randint(INT32_MIN, INT32_MAX, 3).astype(np.int64)

    aux_data = {}

    embedding = (
        10.0
        * (embedding - np.min(embedding, 0))
        / (np.max(embedding, 0) - np.min(embedding, 0))
    ).astype(np.float32, order="C")

    (original_embedding, ghost_embeddings, ghost_mask) = optimize_layout_euclidean(
        n_ghosts,
        embedding,
        head,
        tail,
        n_epochs,
        n_vertices,
        epochs_per_sample,
        a,
        b,
        rng_state,
        gamma,
        initial_alpha,
        negative_sample_rate,
        parallel=parallel,
        verbose=verbose,
        tqdm_kwds=tqdm_kwds,
        move_other=True,
    )

    return original_embedding, ghost_embeddings, ghost_mask, aux_data


class GhostUMAP2(UMAP):
    """GhostUMAP

    GhostUMAP is a variant of UMAP that allows for the embedding of
    additional "ghost" points into the embedding space. These ghost
    points are used to represent the uncertainty in the embedding
    of the original data points.

    Parameters
    ----------
    n_neighbors: float (optional, default 15)
        The size of local neighborhood (in terms of number of neighboring
        sample points) used for manifold approximation. Larger values
        result in more global views of the manifold, while smaller
        values result in more local data being preserved. In general
        values should be in the range 2 to 100.

    n_components: int (optional, default 2)
        The dimension of the space to embed into. This defaults to 2 to
        provide easy visualization, but can reasonably be set to any
        integer value in the range 2 to 100.

    metric: string or function (optional, default 'euclidean')
        The metric to use to compute distances in high dimensional space.
        If a string is passed it must match a valid predefined metric. If
        a general metric is required a function that takes two 1d arrays and
        returns a float can be provided. For performance purposes it is
        required that this be a numba jit'd function. Valid string metrics
        include:

        * euclidean
        * manhattan
        * chebyshev
        * minkowski
        * canberra
        * braycurtis
        * mahalanobis
        * wminkowski
        * seuclidean
        * cosine
        * correlation
        * haversine
        * hamming
        * jaccard
        * dice
        * russelrao
        * kulsinski
        * ll_dirichlet
        * hellinger
        * rogerstanimoto
        * sokalmichener
        * sokalsneath
        * yule

        Metrics that take arguments (such as minkowski, mahalanobis etc.)
        can have arguments passed via the metric_kwds dictionary. At this
        time care must be taken and dictionary elements must be ordered
        appropriately; this will hopefully be fixed in the future.

    n_epochs: int (optional, default None)
        The number of training epochs to be used in optimizing the
        low dimensional embedding. Larger values result in more accurate
        embeddings. If None is specified a value will be selected based on
        the size of the input dataset (200 for large datasets, 500 for small).

    learning_rate: float (optional, default 1.0)
        The initial learning rate for the embedding optimization.

    init: string (optional, default 'spectral')
        How to initialize the low dimensional embedding. Options are:

            * 'spectral': use a spectral embedding of the fuzzy 1-skeleton
            * 'random': assign initial embedding positions at random.
            * 'pca': use the first n_components from PCA applied to the
                input data.
            * 'tswspectral': use a spectral embedding of the fuzzy
                1-skeleton, using a truncated singular value decomposition to
                "warm" up the eigensolver. This is intended as an alternative
                to the 'spectral' method, if that takes an  excessively long
                time to complete initialization (or fails to complete).
            * A numpy array of initial embedding positions.

    min_dist: float (optional, default 0.1)
        The effective minimum distance between embedded points. Smaller values
        will result in a more clustered/clumped embedding where nearby points
        on the manifold are drawn closer together, while larger values will
        result on a more even dispersal of points. The value should be set
        relative to the ``spread`` value, which determines the scale at which
        embedded points will be spread out.

    spread: float (optional, default 1.0)
        The effective scale of embedded points. In combination with ``min_dist``
        this determines how clustered/clumped the embedded points are.

    low_memory: bool (optional, default True)
        For some datasets the nearest neighbor computation can consume a lot of
        memory. If you find that UMAP is failing due to memory constraints
        consider setting this option to True. This approach is more
        computationally expensive, but avoids excessive memory use.

    set_op_mix_ratio: float (optional, default 1.0)
        Interpolate between (fuzzy) union and intersection as the set operation
        used to combine local fuzzy simplicial sets to obtain a global fuzzy
        simplicial sets. Both fuzzy set operations use the product t-norm.
        The value of this parameter should be between 0.0 and 1.0; a value of
        1.0 will use a pure fuzzy union, while 0.0 will use a pure fuzzy
        intersection.

    local_connectivity: int (optional, default 1)
        The local connectivity required -- i.e. the number of nearest
        neighbors that should be assumed to be connected at a local level.
        The higher this value the more connected the manifold becomes
        locally. In practice this should be not more than the local intrinsic
        dimension of the manifold.

    repulsion_strength: float (optional, default 1.0)
        Weighting applied to negative samples in low dimensional embedding
        optimization. Values higher than one will result in greater weight
        being given to negative samples.

    negative_sample_rate: int (optional, default 5)
        The number of negative samples to select per positive sample
        in the optimization process. Increasing this value will result
        in greater repulsive force being applied, greater optimization
        cost, but slightly more accuracy.

    transform_queue_size: float (optional, default 4.0)
        For transform operations (embedding new points using a trained model
        this will control how aggressively to search for nearest neighbors.
        Larger values will result in slower performance but more accurate
        nearest neighbor evaluation.

    a: float (optional, default None)
        More specific parameters controlling the embedding. If None these
        values are set automatically as determined by ``min_dist`` and
        ``spread``.
    b: float (optional, default None)
        More specific parameters controlling the embedding. If None these
        values are set automatically as determined by ``min_dist`` and
        ``spread``.

    random_state: int, RandomState instance or None, optional (default: None)
        If int, random_state is the seed used by the random number generator;
        If RandomState instance, random_state is the random number generator;
        If None, the random number generator is the RandomState instance used
        by `np.random`.

    metric_kwds: dict (optional, default None)
        Arguments to pass on to the metric, such as the ``p`` value for
        Minkowski distance. If None then no arguments are passed on.

    angular_rp_forest: bool (optional, default False)
        Whether to use an angular random projection forest to initialise
        the approximate nearest neighbor search. This can be faster, but is
        mostly only useful for a metric that uses an angular style distance such
        as cosine, correlation etc. In the case of those metrics angular forests
        will be chosen automatically.

    target_n_neighbors: int (optional, default -1)
        The number of nearest neighbors to use to construct the target simplicial
        set. If set to -1 use the ``n_neighbors`` value.

    target_metric: string or callable (optional, default 'categorical')
        The metric used to measure distance for a target array is using supervised
        dimension reduction. By default this is 'categorical' which will measure
        distance in terms of whether categories match or are different. Furthermore,
        if semi-supervised is required target values of -1 will be trated as
        unlabelled under the 'categorical' metric. If the target array takes
        continuous values (e.g. for a regression problem) then metric of 'l1'
        or 'l2' is probably more appropriate.

    target_metric_kwds: dict (optional, default None)
        Keyword argument to pass to the target metric when performing
        supervised dimension reduction. If None then no arguments are passed on.

    target_weight: float (optional, default 0.5)
        weighting factor between data topology and target topology. A value of
        0.0 weights predominantly on data, a value of 1.0 places a strong emphasis on
        target. The default of 0.5 balances the weighting equally between data and
        target.

    transform_seed: int (optional, default 42)
        Random seed used for the stochastic aspects of the transform operation.
        This ensures consistency in transform operations.

    verbose: bool (optional, default False)
        Controls verbosity of logging.

    tqdm_kwds: dict (optional, defaul None)
        Key word arguments to be used by the tqdm progress bar.

    unique: bool (optional, default False)
        Controls if the rows of your data should be uniqued before being
        embedded.  If you have more duplicates than you have ``n_neighbors``
        you can have the identical data points lying in different regions of
        your space.  It also violates the definition of a metric.
        For to map from internal structures back to your data use the variable
        _unique_inverse_.

    densmap: bool (optional, default False)
        Specifies whether the density-augmented objective of densMAP
        should be used for optimization. Turning on this option generates
        an embedding where the local densities are encouraged to be correlated
        with those in the original space. Parameters below with the prefix 'dens'
        further control the behavior of this extension.

    dens_lambda: float (optional, default 2.0)
        Controls the regularization weight of the density correlation term
        in densMAP. Higher values prioritize density preservation over the
        UMAP objective, and vice versa for values closer to zero. Setting this
        parameter to zero is equivalent to running the original UMAP algorithm.

    dens_frac: float (optional, default 0.3)
        Controls the fraction of epochs (between 0 and 1) where the
        density-augmented objective is used in densMAP. The first
        (1 - dens_frac) fraction of epochs optimize the original UMAP objective
        before introducing the density correlation term.

    dens_var_shift: float (optional, default 0.1)
        A small constant added to the variance of local radii in the
        embedding when calculating the density correlation objective to
        prevent numerical instability from dividing by a small number

    output_dens: float (optional, default False)
        Determines whether the local radii of the final embedding (an inverse
        measure of local density) are computed and returned in addition to
        the embedding. If set to True, local radii of the original data
        are also included in the output for comparison; the output is a tuple
        (embedding, original local radii, embedding local radii). This option
        can also be used when densmap=False to calculate the densities for
        UMAP embeddings.

    disconnection_distance: float (optional, default np.inf or maximal value for bounded distances)
        Disconnect any vertices of distance greater than or equal to disconnection_distance when approximating the
        manifold via our k-nn graph. This is particularly useful in the case that you have a bounded metric.  The
        UMAP assumption that we have a connected manifold can be problematic when you have points that are maximally
        different from all the rest of your data.  The connected manifold assumption will make such points have perfect
        similarity to a random set of other points.  Too many such points will artificially connect your space.

    precomputed_knn: tuple (optional, default (None,None,None))
        If the k-nearest neighbors of each point has already been calculated you
        can pass them in here to save computation time. The number of nearest
        neighbors in the precomputed_knn must be greater or equal to the
        n_neighbors parameter. This should be a tuple containing the output
        of the nearest_neighbors() function or attributes from a previously fit
        UMAP object; (knn_indices, knn_dists, knn_search_index). If you wish to use
        k-nearest neighbors data calculated by another package then provide a tuple of
        the form (knn_indices, knn_dists). The contents of the tuple should be two numpy
        arrays of shape (N, n_neighbors) where N is the number of items in the
        input data. The first array should be the integer indices of the nearest
        neighbors, and the second array should be the corresponding distances. The
        nearest neighbor of each item should be itself, e.g. the nearest neighbor of
        item 0 should be 0, the nearest neighbor of item 1 is 1 and so on. Please note
        that you will *not* be able to transform new data in this case.
    """

    def __init__(self, *args, **kwargs):
        super().__init__(*args, **kwargs)

    def _fit_embed_data(self, X, n_ghosts, n_epochs, init, random_state):
        """A method wrapper for simplicial_set_embedding_with_ghost that can be
        replaced by subclasses.
        """
        return simplicial_set_embedding(
            X,
            n_ghosts,
            self.graph_,
            self.n_components,
            self._initial_alpha,
            self._a,
            self._b,
            self.repulsion_strength,
            self.negative_sample_rate,
            n_epochs,
            init,
            random_state,
            self._input_distance_func,
            self._metric_kwds,
            self.random_state is None,
            self.verbose,
            tqdm_kwds=self.tqdm_kwds,
        )

    def fit(
        self,
        X,
        y=None,
        force_all_finite=True,
        n_ghosts=16,
    ):
        """Fit X into an embedded space.

        Optionally use y for supervised dimension reduction.

        Parameters
        ----------
        X : array, shape (n_samples, n_features) or (n_samples, n_samples)
            If the metric is 'precomputed' X must be a square distance
            matrix. Otherwise it contains a sample per row. If the method
            is 'exact', X may be a sparse matrix of type 'csr', 'csc'
            or 'coo'.

        y : array, shape (n_samples)
            A target array for supervised dimension reduction. How this is
            handled is determined by parameters UMAP was instantiated with.
            The relevant attributes are ``target_metric`` and
            ``target_metric_kwds``.

        force_all_finite : Whether to raise an error on np.inf, np.nan, pd.NA in array.
            The possibilities are: - True: Force all values of array to be finite.
                                   - False: accepts np.inf, np.nan, pd.NA in array.
                                   - 'allow-nan': accepts only np.nan and pd.NA values in array.
                                     Values cannot be infinite.
        """

        X = check_array(
            X,
            dtype=np.float32,
            accept_sparse="csr",
            order="C",
            force_all_finite=force_all_finite,
        )
        self._raw_data = X

        # Handle all the optional arguments, setting default
        if self.a is None or self.b is None:
            self._a, self._b = find_ab_params(self.spread, self.min_dist)
        else:
            self._a = self.a
            self._b = self.b

        if isinstance(self.init, np.ndarray):
            init = check_array(
                self.init,
                dtype=np.float32,
                accept_sparse=False,
                force_all_finite=force_all_finite,
            )
        else:
            init = self.init

        self._initial_alpha = self.learning_rate

        self.knn_indices = self.precomputed_knn[0]
        self.knn_dists = self.precomputed_knn[1]
        # #848: allow precomputed knn to not have a search index
        if len(self.precomputed_knn) == 2:
            self.knn_search_index = None
        else:
            self.knn_search_index = self.precomputed_knn[2]

        self._validate_parameters()

        if self.verbose:
            print(str(self))

        self._original_n_threads = numba.get_num_threads()
        if self.n_jobs > 0 and self.n_jobs is not None:
            numba.set_num_threads(self.n_jobs)

        # Check if we should unique the data
        # We've already ensured that we aren't in the precomputed case
        if self.unique:
            # check if the matrix is dense
            if self._sparse_data:
                # Call a sparse unique function
                index, inverse, counts = csr_unique(X)
            else:
                index, inverse, counts = np.unique(
                    X,
                    return_index=True,
                    return_inverse=True,
                    return_counts=True,
                    axis=0,
                )[1:4]
            if self.verbose:
                print(
                    "Unique=True -> Number of data points reduced from ",
                    X.shape[0],
                    " to ",
                    X[index].shape[0],
                )
                most_common = np.argmax(counts)
                print(
                    "Most common duplicate is",
                    index[most_common],
                    " with a count of ",
                    counts[most_common],
                )
            # We'll expose an inverse map when unique=True for users to map from our internal structures to their data
            self._unique_inverse_ = inverse
        # If we aren't asking for unique use the full index.
        # This will save special cases later.
        else:
            index = list(range(X.shape[0]))
            inverse = list(range(X.shape[0]))

        # Error check n_neighbors based on data size
        if X[index].shape[0] <= self.n_neighbors:
            if X[index].shape[0] == 1:
                self.embedding_ = np.zeros(
                    (1, self.n_components)
                )  # needed to sklearn comparability
                return self

            warn(
                "n_neighbors is larger than the dataset size; truncating to "
                "X.shape[0] - 1"
            )
            self._n_neighbors = X[index].shape[0] - 1
            if self.densmap:
                self._densmap_kwds["n_neighbors"] = self._n_neighbors
        else:
            self._n_neighbors = self.n_neighbors

        # Note: unless it causes issues for setting 'index', could move this to
        # initial sparsity check above
        if self._sparse_data and not X.has_sorted_indices:
            X.sort_indices()

        random_state = check_random_state(self.random_state)

        if self.verbose:
            print(ts(), "Construct fuzzy simplicial set")

        if self.metric == "precomputed" and self._sparse_data:
            # For sparse precomputed distance matrices, we just argsort the rows to find
            # nearest neighbors. To make this easier, we expect matrices that are
            # symmetrical (so we can find neighbors by looking at rows in isolation,
            # rather than also having to consider that sample's column too).
            # print("Computing KNNs for sparse precomputed distances...")
            if sparse_tril(X).getnnz() != sparse_triu(X).getnnz():
                raise ValueError(
                    "Sparse precomputed distance matrices should be symmetrical!"
                )
            if not np.all(X.diagonal() == 0):
                raise ValueError("Non-zero distances from samples to themselves!")
            if self.knn_dists is None:
                self._knn_indices = np.zeros((X.shape[0], self.n_neighbors), dtype=int)
                self._knn_dists = np.zeros(self._knn_indices.shape, dtype=float)
                for row_id in range(X.shape[0]):
                    # Find KNNs row-by-row
                    row_data = X[row_id].data
                    row_indices = X[row_id].indices
                    if len(row_data) < self._n_neighbors:
                        raise ValueError(
                            "Some rows contain fewer than n_neighbors distances!"
                        )
                    row_nn_data_indices = np.argsort(row_data)[: self._n_neighbors]
                    self._knn_indices[row_id] = row_indices[row_nn_data_indices]
                    self._knn_dists[row_id] = row_data[row_nn_data_indices]
            else:
                self._knn_indices = self.knn_indices
                self._knn_dists = self.knn_dists
            # Disconnect any vertices farther apart than _disconnection_distance
            disconnected_index = self._knn_dists >= self._disconnection_distance
            self._knn_indices[disconnected_index] = -1
            self._knn_dists[disconnected_index] = np.inf
            edges_removed = disconnected_index.sum()

            (
                self.graph_,
                self._sigmas,
                self._rhos,
                self.graph_dists_,
            ) = fuzzy_simplicial_set(
                X[index],
                self.n_neighbors,
                random_state,
                "precomputed",
                self._metric_kwds,
                self._knn_indices,
                self._knn_dists,
                self.angular_rp_forest,
                self.set_op_mix_ratio,
                self.local_connectivity,
                True,
                self.verbose,
                self.densmap or self.output_dens,
            )
            # Report the number of vertices with degree 0 in our our umap.graph_
            # This ensures that they were properly disconnected.
            vertices_disconnected = np.sum(
                np.array(self.graph_.sum(axis=1)).flatten() == 0
            )
            raise_disconnected_warning(
                edges_removed,
                vertices_disconnected,
                self._disconnection_distance,
                self._raw_data.shape[0],
                verbose=self.verbose,
            )
        # Handle small cases efficiently by computing all distances
        elif X[index].shape[0] < 4096 and not self.force_approximation_algorithm:
            self._small_data = True
            try:
                # sklearn pairwise_distances fails for callable metric on sparse data
                _m = self.metric if self._sparse_data else self._input_distance_func
                dmat = pairwise_distances(X[index], metric=_m, **self._metric_kwds)
            except (ValueError, TypeError) as e:
                # metric is numba.jit'd or not supported by sklearn,
                # fallback to pairwise special

                if self._sparse_data:
                    # Get a fresh metric since we are casting to dense
                    if not callable(self.metric):
                        _m = dist.named_distances[self.metric]
                        dmat = dist.pairwise_special_metric(
                            X[index].toarray(),
                            metric=_m,
                            kwds=self._metric_kwds,
                            force_all_finite=force_all_finite,
                        )
                    else:
                        dmat = dist.pairwise_special_metric(
                            X[index],
                            metric=self._input_distance_func,
                            kwds=self._metric_kwds,
                            force_all_finite=force_all_finite,
                        )
                else:
                    dmat = dist.pairwise_special_metric(
                        X[index],
                        metric=self._input_distance_func,
                        kwds=self._metric_kwds,
                        force_all_finite=force_all_finite,
                    )
            # set any values greater than disconnection_distance to be np.inf.
            # This will have no effect when _disconnection_distance is not set since it defaults to np.inf.
            edges_removed = np.sum(dmat >= self._disconnection_distance)
            dmat[dmat >= self._disconnection_distance] = np.inf
            (
                self.graph_,
                self._sigmas,
                self._rhos,
                self.graph_dists_,
            ) = fuzzy_simplicial_set(
                dmat,
                self._n_neighbors,
                random_state,
                "precomputed",
                self._metric_kwds,
                None,
                None,
                self.angular_rp_forest,
                self.set_op_mix_ratio,
                self.local_connectivity,
                True,
                self.verbose,
                self.densmap or self.output_dens,
            )
            # Report the number of vertices with degree 0 in our umap.graph_
            # This ensures that they were properly disconnected.
            vertices_disconnected = np.sum(
                np.array(self.graph_.sum(axis=1)).flatten() == 0
            )
            raise_disconnected_warning(
                edges_removed,
                vertices_disconnected,
                self._disconnection_distance,
                self._raw_data.shape[0],
                verbose=self.verbose,
            )
        else:
            # Standard case
            self._small_data = False
            # Standard case
            if self._sparse_data and self.metric in pynn_sparse_named_distances:
                nn_metric = self.metric
            elif not self._sparse_data and self.metric in pynn_named_distances:
                nn_metric = self.metric
            else:
                nn_metric = self._input_distance_func
            if self.knn_dists is None:
                (
                    self._knn_indices,
                    self._knn_dists,
                    self._knn_search_index,
                ) = nearest_neighbors(
                    X[index],
                    self._n_neighbors,
                    nn_metric,
                    self._metric_kwds,
                    self.angular_rp_forest,
                    random_state,
                    self.low_memory,
                    use_pynndescent=True,
                    n_jobs=self.n_jobs,
                    verbose=self.verbose,
                )
                self.knn_indices = self._knn_indices
                self.knn_dists = self._knn_dists
                self.knn_search_index = self._knn_search_index
            else:
                self._knn_indices = self.knn_indices
                self._knn_dists = self.knn_dists
                self._knn_search_index = self.knn_search_index
            # Disconnect any vertices farther apart than _disconnection_distance
            disconnected_index = self._knn_dists >= self._disconnection_distance
            self._knn_indices[disconnected_index] = -1
            self._knn_dists[disconnected_index] = np.inf
            edges_removed = disconnected_index.sum()

            (
                self.graph_,
                self._sigmas,
                self._rhos,
                self.graph_dists_,
            ) = fuzzy_simplicial_set(
                X[index],
                self.n_neighbors,
                random_state,
                nn_metric,
                self._metric_kwds,
                self._knn_indices,
                self._knn_dists,
                self.angular_rp_forest,
                self.set_op_mix_ratio,
                self.local_connectivity,
                True,
                self.verbose,
                self.densmap or self.output_dens,
            )
            # Report the number of vertices with degree 0 in our umap.graph_
            # This ensures that they were properly disconnected.
            vertices_disconnected = np.sum(
                np.array(self.graph_.sum(axis=1)).flatten() == 0
            )
            raise_disconnected_warning(
                edges_removed,
                vertices_disconnected,
                self._disconnection_distance,
                self._raw_data.shape[0],
                verbose=self.verbose,
            )

        # Currently not checking if any duplicate points have differing labels
        # Might be worth throwing a warning...
        if y is not None:
            len_X = len(X) if not self._sparse_data else X.shape[0]
            if len_X != len(y):
                raise ValueError(
                    "Length of x = {len_x}, length of y = {len_y}, while it must be equal.".format(
                        len_x=len_X, len_y=len(y)
                    )
                )
            if self.target_metric == "string":
                y_ = y[index]
            else:
                y_ = check_array(y, ensure_2d=False, force_all_finite=force_all_finite)[
                    index
                ]
            if self.target_metric == "categorical":
                if self.target_weight < 1.0:
                    far_dist = 2.5 * (1.0 / (1.0 - self.target_weight))
                else:
                    far_dist = 1.0e12
                self.graph_ = discrete_metric_simplicial_set_intersection(
                    self.graph_, y_, far_dist=far_dist
                )
            elif self.target_metric in dist.DISCRETE_METRICS:
                if self.target_weight < 1.0:
                    scale = 2.5 * (1.0 / (1.0 - self.target_weight))
                else:
                    scale = 1.0e12
                # self.graph_ = discrete_metric_simplicial_set_intersection(
                #     self.graph_,
                #     y_,
                #     metric=self.target_metric,
                #     metric_kws=self.target_metric_kwds,
                #     metric_scale=scale
                # )

                metric_kws = dist.get_discrete_params(y_, self.target_metric)

                self.graph_ = discrete_metric_simplicial_set_intersection(
                    self.graph_,
                    y_,
                    metric=self.target_metric,
                    metric_kws=metric_kws,
                    metric_scale=scale,
                )
            else:
                if len(y_.shape) == 1:
                    y_ = y_.reshape(-1, 1)
                if self.target_n_neighbors == -1:
                    target_n_neighbors = self._n_neighbors
                else:
                    target_n_neighbors = self.target_n_neighbors

                # Handle the small case as precomputed as before
                if y.shape[0] < 4096:
                    try:
                        ydmat = pairwise_distances(
                            y_, metric=self.target_metric, **self._target_metric_kwds
                        )
                    except (TypeError, ValueError):
                        ydmat = dist.pairwise_special_metric(
                            y_,
                            metric=self.target_metric,
                            kwds=self._target_metric_kwds,
                            force_all_finite=force_all_finite,
                        )

                    (
                        target_graph,
                        target_sigmas,
                        target_rhos,
                    ) = fuzzy_simplicial_set(
                        ydmat,
                        target_n_neighbors,
                        random_state,
                        "precomputed",
                        self._target_metric_kwds,
                        None,
                        None,
                        False,
                        1.0,
                        1.0,
                        False,
                    )
                else:
                    # Standard case
                    (
                        target_graph,
                        target_sigmas,
                        target_rhos,
                    ) = fuzzy_simplicial_set(
                        y_,
                        target_n_neighbors,
                        random_state,
                        self.target_metric,
                        self._target_metric_kwds,
                        None,
                        None,
                        False,
                        1.0,
                        1.0,
                        False,
                    )
                # product = self.graph_.multiply(target_graph)
                # # self.graph_ = 0.99 * product + 0.01 * (self.graph_ +
                # #                                        target_graph -
                # #                                        product)
                # self.graph_ = product
                self.graph_ = general_simplicial_set_intersection(
                    self.graph_, target_graph, self.target_weight
                )
                self.graph_ = reset_local_connectivity(self.graph_)
            self._supervised = True
        else:
            self._supervised = False

        if self.densmap or self.output_dens:
            self._densmap_kwds["graph_dists"] = self.graph_dists_

        if self.verbose:
            print(ts(), "Construct embedding")

        if self.transform_mode == "embedding":
            epochs = (
                self.n_epochs_list if self.n_epochs_list is not None else self.n_epochs
            )
            (
                self.original_embedding,
                self.ghost_embeddings,
                self.ghost_mask,
                aux_data,
            ) = self._fit_embed_data(
                self._raw_data[index],
                n_ghosts,
                epochs,
                init,
                random_state,
            )
            if self.n_epochs_list is not None:
                if "embedding_list" not in aux_data:
                    raise KeyError(
                        "No list of embedding were found in 'aux_data'. "
                        "It is likely the layout optimization function "
                        "doesn't support the list of int for 'n_epochs'."
                    )
                else:
                    self.embedding_list_ = [
                        e[inverse] for e in aux_data["embedding_list"]
                    ]

            # Assign any points that are fully disconnected from our manifold(s) to have embedding
            # coordinates of np.nan.  These will be filtered by our plotting functions automatically.
            # They also prevent users from being deceived a distance query to one of these points.
            # Might be worth moving this into simplicial_set_embedding or _fit_embed_data
            disconnected_vertices = np.array(self.graph_.sum(axis=1)).flatten() == 0
            # if len(disconnected_vertices) > 0:
            #     self.embedding_[disconnected_vertices] = np.full(
            #         self.n_components, np.nan
            #     )

            # self.embedding_ = self.embedding_[inverse]
            # if self.output_dens:
            #     self.rad_orig_ = aux_data["rad_orig"][inverse]
            #     self.rad_emb_ = aux_data["rad_emb"][inverse]

        if self.verbose:
            print(ts() + " Finished embedding")

        numba.set_num_threads(self._original_n_threads)
        self._input_hash = joblib.hash(self._raw_data)

        return self

    def fit_transform(
        self,
        X: np.ndarray,
        force_all_finite: bool = True,
        n_ghosts: int = 8,
        r: float = 0.1,
        sensitivity: float = 1,
        ghost_gen: float = 0.2,
        dropping: bool = True,
        init_dropping: float = 0.4,
        bm_type: str = "None",
    ):
        """
        Fit X into an embedded space with ghosts and return that transformed outputs.

        Parameters
        ----------
        X : array, shape (n_samples, n_features) or (n_samples, n_samples)
            If the metric is 'precomputed' X must be a square distance
            matrix. Otherwise it contains a sample per row. If the method
            is 'exact', X may be a sparse matrix of type 'csr', 'csc'
            or 'coo'.

        force_all_finite : Whether to raise an error on np.inf, np.nan, pd.NA in array.
            The possibilities are: - True: Force all values of array to be finite.
                                    - False: accepts np.inf, np.nan, pd.NA in array.
                                    - 'allow-nan': accepts only np.nan and pd.NA values in array.
                                       Values cannot be infinite.

        n_ghosts : int (optional, default 8)
            The number of ghost points to embed in the embedding space.

        schedule : list of int (optional, default None)
            The schedule of epochs to successive halving.

        Returns
        -------
        embedding : array, shape (n_samples, n_components)
            The transformed samples in the embedded space.

        ghost_embedding : array, shape (n_samples, n_ghosts, n_components)
            The transformed ghost points in the embedded space.

        ghost_indices : array, shape (n_remaining_ghosts,)
            The indices of the ghost points in the original data.

        """
        set_config(
            r=r,
            sensitivity=sensitivity,
            ghost_gen=ghost_gen,
            dropping=dropping,
            init_dropping=init_dropping,
            smoothing_factor=0.9,
            bm_type=bm_type,
        )

        if n_ghosts < 1:
            raise ValueError("n_ghosts should be greater than 0")
        self.radius = r
        y = None
        self.fit(X, y, force_all_finite, n_ghosts)

        return (self.original_embedding, self.ghost_embeddings, self.ghost_mask)

    def get_radii(self) -> np.ndarray:
        if not hasattr(self, "original_embedding"):
            raise ValueError("The model has not been fitted yet.")

        return compute_distances(self.original_embedding, self.ghost_embeddings)

    def get_unstable_ghosts(
        self, distance: float = 0.1, sensitivity: float = 1
    ) -> np.ndarray:
        if not hasattr(self, "original_embedding"):
            raise ValueError("The model has not been fitted yet.")

        ghost_mask = np.ones(self.original_embedding.shape[0], dtype=bool)

        ghost_mask = drop_ghosts(
            self.original_embedding,
            self.ghost_embeddings,
            ghost_mask,
            distance=distance,
            sensitivity=sensitivity,
        )

        return ghost_mask

    def get_distances(self, sensitivity: float = 1) -> np.ndarray:
        return get_distance(
            self.original_embedding,
            self.ghost_embeddings,
            self.ghost_mask,
            sensitivity,
        )

    def get_results(self):
        return _get_results()

    def get_config(self):
        return _get_config()

    def visualize(self, label=None, legend=None):
        """Creates and returns an interactive visualization widget.

        The widget will be automatically displayed in Jupyter environments.
        """

        init_radii = _get_results().init_radii
        widget = Widget(
            original_embedding=self.original_embedding,
            ghost_embedding=self.ghost_embeddings,
            r=self.radius,
            init_radii=init_radii,
            neighbors=(
                self._knn_indices
                if hasattr(self, "_knn_indices")
                else self.get_NN_for_small_data()
            ),
            label=label,
            legend=legend,
        )
        return widget

    def get_NN_for_small_data(self):
        from sklearn.neighbors import NearestNeighbors

        nn = NearestNeighbors(n_neighbors=self._n_neighbors, metric=self.metric)

        nn.fit(self._raw_data)

        _, indices = nn.kneighbors(self._raw_data)

        return indices



import os
import json

from ..BaseLoader import BaseLoader

import numpy as np

from umap import UMAP


class Cnae9Loader(BaseLoader):
    def __init__(self):
        super().__init__()
        self.name = "cnae9"
        self.base_path = os.path.dirname(os.path.abspath(__file__))
        self.load_data()

    def load_raw_data(self):
        import pandas as pd

        data = pd.read_csv(os.path.join(self.base_path, "CNAE-9.data"), header=None)

        X = data.values[:, 1:]
        y = data.values[:, 0]

        self._data = self.scale_data(X)
        self._label = y
        self._legend = [f"{i}" for i in range(10)]
        self._precomputed_knn = self.get_precomputed_knn(self._data)

        self.save_data(self.base_path)

    def get_data(self):
        return super().get_data()



import numpy as np


def sample_ghosts(
    original_embedding: np.ndarray,
    n_samples: int,
    r: float = 0.1,
) -> np.ndarray:
    """
    Sample ghost projections around the original projections within a radius r
    """

    max_dist = _get_max_extent(original_embedding)

    r1, r2 = (0, r * max_dist)

    n_points = original_embedding.shape[0]

    theta = np.random.uniform(0, 2 * np.pi, (n_points, n_samples))
    radii = np.random.uniform(r1, r2, (n_points, n_samples))

    dx = radii * np.cos(theta)
    dy = radii * np.sin(theta)

    samples = np.empty((n_points, n_samples, 2))
    samples = original_embedding[:, np.newaxis, :] + np.stack([dx, dy], axis=-1)

    return np.array(samples, dtype=np.float32, order="C")


def drop_ghosts(
    original_embedding: np.ndarray,
    ghost_embeddings: np.ndarray,
    ghost_mask: np.ndarray,
    sensitivity: float = 1,
    distance: int = 0.1,
) -> np.ndarray:
    """
    Drop stable ghosts based on the distance
    """

    active_indices = np.where(ghost_mask)[0]

    distances = get_distance(
        original_embedding, ghost_embeddings, ghost_mask, sensitivity
    )

    dropped_idx = np.where(distances < distance)[0]

    ghost_mask[active_indices[dropped_idx]] = False

    return ghost_mask


def get_distance(
    original_embedding: np.ndarray,
    ghost_embeddings: np.ndarray,
    ghost_mask: np.ndarray,
    sensitivity: float = 1,
) -> np.ndarray:
    """
    Compute the distance d of each data point
    """

    max_extent = _get_max_extent(original_embedding)
    boundary = np.ceil((ghost_embeddings.shape[1] - 1) * sensitivity).astype(int)

    distances = compute_distances(
        original_embedding[ghost_mask], ghost_embeddings[ghost_mask]
    )
    distances = np.sort(distances, axis=1)
    D = distances[:, boundary] / max_extent

    return D


def _get_max_extent(original_embedding: np.ndarray) -> np.ndarray:
    """
    Compute the maximum extent of the embedding
    """

    x_range = np.max(original_embedding[:, 0]) - np.min(original_embedding[:, 0])
    y_range = np.max(original_embedding[:, 1]) - np.min(original_embedding[:, 1])
    max_extent = np.max([x_range, y_range])

    return max_extent


def compute_distances(
    original_embedding: np.ndarray, ghost_embeddings: np.ndarray
) -> np.ndarray:
    """
    Compute the distances between the original projection and its ghost projections
    """

    distances = np.sum(
        (ghost_embeddings - original_embedding[:, np.newaxis, :]) ** 2, axis=2
    )
    distances **= 0.5

    return distances


def _measure_instability(
    original_embedding: np.ndarray,
    ghost_embeddings: np.ndarray,
    ghost_mask: np.ndarray,
):
    O = original_embedding[ghost_mask]
    G = ghost_embeddings[ghost_mask]

    Y = np.concatenate([O[:, np.newaxis], G], axis=1)

    Mu = np.mean(Y, axis=1)

    INS = np.sum(np.square(Y - Mu[:, np.newaxis]), axis=2)
    INS = np.mean(INS, axis=1)

    rank = np.argsort(INS)[::-1]
    score = INS[rank]

    rank = np.where(ghost_mask)[0][rank]

    return rank, score



import importlib.metadata
import pathlib
from typing import List, TypedDict

import anywidget
import traitlets
import numpy as np
import os

from .model import EmbeddingSet


class EmbeddingSetInput(TypedDict):
    original_embedding: np.ndarray
    ghost_embedding: np.ndarray
    r: float
    init_radii: np.ndarray
    neighbors: np.ndarray | list[list[int]]
    title: str | None
    legend: list[str] | None
    colors: dict[str, str] | None


try:
    __version__ = importlib.metadata.version("ghostumap")
except importlib.metadata.PackageNotFoundError:
    __version__ = "unknown"


class Widget(anywidget.AnyWidget):
    _esm = (
        pathlib.Path(__file__).parent / "static" / "widget.js"
        if not os.getenv("ANYWIDGET_HMR")
        else "http://localhost:5173/widget/widget.ts?anywidget"
    )
    _css_path = pathlib.Path(__file__).parent / "static" / "widget.css"
    _css = _css_path if _css_path.exists() else None

    width = traitlets.Int(300).tag(sync=True)
    height = traitlets.Int(300).tag(sync=True)
    legend_width = traitlets.Int(150).tag(sync=True)
    legend_height = traitlets.Int(300).tag(sync=True)
    histogram_width = traitlets.Int(300).tag(sync=True)
    histogram_height = traitlets.Int(300).tag(sync=True)

    embedding_set: list[EmbeddingSet] = traitlets.List([]).tag(sync=True)

    r = traitlets.Float(0.1).tag(sync=True)
    distance = traitlets.Float(0.1).tag(sync=True)
    sensitivity = traitlets.Float(0.9).tag(sync=True)
    show_neighbors = traitlets.Bool(True).tag(sync=False)
    show_ghosts = traitlets.Bool(True).tag(sync=True)
    show_unstables = traitlets.Bool(True).tag(sync=True)
    embedding_id = traitlets.Int(0).tag(sync=True)

    def __init__(
        self,
        embedding_set: list[EmbeddingSetInput] | EmbeddingSetInput | None = None,
        original_embedding: np.ndarray = None,
        ghost_embedding: np.ndarray = None,
        r: float = 0.1,
        init_radii: np.ndarray = None,
        neighbors: np.ndarray = None,
        label: np.ndarray | list[str] = None,
        title: str = None,
        legend: list[str] = None,
        sensitivity: float = 0.9,
        distance: float = 0.1,
        width: int = 600,
        height: int = 600,
        legend_width: int = 300,
        legend_height: int = 600,
        *args,
        **kwargs,
    ):
        super().__init__(*args, **kwargs)

        self.width = width
        self.height = height
        self.legend_width = legend_width
        self.legend_height = legend_height

        self.sensitivity = sensitivity
        self.distance = distance
        self.r = r
        self.init_radii = init_radii

        self.embedding_set = self._process_embedding(
            embedding=embedding_set,
            original_embedding=original_embedding,
            ghost_embedding=ghost_embedding,
            r=r,
            init_radii=init_radii,
            neighbors=neighbors,
            label=label,
            title=title,
            legend=legend,
        )

    def add_embedding(
        self,
        embedding: list[EmbeddingSetInput] | EmbeddingSetInput | None = None,
        original_embedding: np.ndarray = None,
        ghost_embedding: np.ndarray = None,
        r: float = 0.1,
        init_radii: np.ndarray = None,
        neighbors: np.ndarray = None,
        label: np.ndarray | list[str] = None,
        title: str = None,
        legend: list[str] = None,
    ):
        self.embedding_set.append(
            self._process_embedding(
                embedding=embedding,
                original_embedding=original_embedding,
                ghost_embedding=ghost_embedding,
                r=r,
                init_radii=init_radii,
                neighbors=neighbors,
                label=label,
                title=title,
                legend=legend,
            )
        )

    def _process_embedding(
        self,
        embedding: list[EmbeddingSetInput] | EmbeddingSetInput | None = None,
        original_embedding: np.ndarray | None = None,
        ghost_embedding: np.ndarray | None = None,
        r: float = 0.1,
        init_radii: np.ndarray = None,
        neighbors: np.ndarray | list[list[int]] | None = None,
        label: np.ndarray | list[str] | None = None,
        title: str | None = None,
        legend: list[str] | None = None,
        colors: dict[str, str] | None = None,
    ):
        if isinstance(embedding, dict):

            return [
                EmbeddingSet(
                    original_embedding=embedding["original_embedding"],
                    ghost_embedding=embedding["ghost_embedding"],
                    r=embedding.get("r"),
                    init_radii=embedding.get("init_radii"),
                    neighbors=embedding["neighbors"],
                    label=embedding.get("label"),
                    title=embedding.get("title"),
                    legend=embedding.get("legend"),
                    colors=embedding.get("colors"),
                ).to_dict()
            ]

        elif isinstance(embedding, list):
            return [
                EmbeddingSet(
                    original_embedding=emb["original_embedding"],
                    ghost_embedding=emb["ghost_embedding"],
                    r=emb.get("r"),
                    init_radii=emb.get("init_radii"),
                    neighbors=emb["neighbors"],
                    label=emb.get("label"),
                    title=emb.get("title"),
                    legend=emb.get("legend"),
                    colors=emb.get("colors"),
                ).to_dict()
                for emb in embedding
            ]

        elif (
            original_embedding is not None
            and ghost_embedding is not None
            and neighbors is not None
        ):
            return [
                EmbeddingSet(
                    original_embedding=original_embedding,
                    ghost_embedding=ghost_embedding,
                    r=r,
                    init_radii=init_radii,
                    neighbors=neighbors,
                    label=label,
                    title=title,
                    legend=legend,
                    colors=colors,
                ).to_dict()
            ]

        else:
            raise ValueError(
                "Either 'embedding' or 'original_embedding' and 'ghost_embedding' must be provided"
            )

    def update_params(
        self,
        **kwargs,
    ):
        """
        Update widget parameters based on provided keyword arguments.

        Parameters:
        - sensitivity: float
        - distance: float
        - show_neighbors: bool
        - show_ghosts: bool
        - width: int
        - height: int
        - legend_width: int
        - legend_height: int

        Any parameter not provided will retain its current value.
        """

        valid_params = {
            "sensitivity",
            "distance",
            "show_neighbors",
            "show_ghosts",
            "width",
            "height",
            "legend_width",
            "legend_height",
        }

        invalid_params = set(kwargs.keys()) - valid_params

        if invalid_params:
            raise ValueError(f"Invalid parameter(s): {', '.join(invalid_params)}")

        for key, value in kwargs.items():
            if value is None:
                raise ValueError(f"Parameter '{key}' cannot be None")
            setattr(self, key, value)

    def _repr_html_(self):
        """IPython display representation"""
        mimebundle = self._repr_mimebundle_()
        if "text/html" in mimebundle[0]:
            return mimebundle[0]["text/html"]
        else:
            return repr(self)



from dataclasses import dataclass

import numpy as np

from ghostumap.utils import compute_distances as compute_distance


@dataclass
class OriginalPointModel:
    id: int
    x: float
    y: float
    radii: list[float]
    label: str | None
    neighbors: list[int]

    def to_dict(self):
        return {
            "id": self.id,
            "x": self.x,
            "y": self.y,
            "radii": self.radii,
            "instability": np.var(self.radii),
            "label": self.label,
            "neighbors": self.neighbors,
        }


class OriginalEmbedding:
    @staticmethod
    def build_model(
        original_embedding: np.ndarray,
        ghost_embedding: np.ndarray,
        neighbors: np.ndarray,
        label: np.ndarray | list[str] | None = None,
    ):
        if label is None:
            label = ["None"] * len(original_embedding)

        radii = compute_distance(original_embedding, ghost_embedding)
        radii = np.sort(radii, axis=1)

        return [
            OriginalPointModel(
                id=i, x=x, y=y, radii=r.tolist(), label=str(l), neighbors=n.tolist()
            ).to_dict()
            for i, (x, y, r, l, n) in enumerate(
                zip(
                    original_embedding[:, 0],
                    original_embedding[:, 1],
                    radii,
                    label,
                    neighbors,
                )
            )
        ]



import logging

import numpy as np

from benchmark.hyperparameters import generate_hyperparameter_comb
from benchmark.utils import run_GhostUMAP, measure_accuracy
from benchmark.save_manager import save_embeddings, save_results
from rdumap.data import DataLoader
from rdumap.ghostumap.utils import drop_ghosts


def benchmark_v0(
    data_name: str,
    X: np.ndarray,
    precomputed_knn,
    hyperparameters,
    iterations=10,
    # data_name: str, base_settings: dict, param_grid: dict, iterations: int = 10
):
    """
    Testing the best hyperparameter combinations for ghost_gen and init_dropping
    """

    # Load dataset

    results = []

    for i in range(iterations):
        logging.info(f"Starting iteration {i + 1}/{iterations}.")
        result_acc = run_GhostUMAP(
            X,
            hyperparameters,
            bm_type="accuracy",
            precomputed_knn=precomputed_knn,
            distance=0.1,
        )
        save_embeddings(result_acc, hyperparameters, "accuracy", data_name)

        result_twd = run_GhostUMAP(
            X,
            hyperparameters,
            bm_type="time_with_dropping",
            precomputed_knn=precomputed_knn,
        )
        save_embeddings(result_twd, hyperparameters, "time_with_dropping", data_name)

        result_twod = run_GhostUMAP(
            X,
            hyperparameters,
            bm_type="time_without_dropping",
            precomputed_knn=precomputed_knn,
        )
        save_embeddings(
            result_twod, hyperparameters, "time_without_dropping", data_name
        )

        f1, precision, recall = measure_accuracy(
            result_acc["unstable_ghosts"], result_acc["alive_ghosts"]
        )
        time_with_dropping = result_twd["opt_time"]
        time_without_dropping = result_twod["opt_time"]

        result = {
            "data": data_name,
            **hyperparameters,
            "iter": i,
            "f1": f1,
            "precision": precision,
            "recall": recall,
            "num_unstable_ghosts": np.sum(result_acc["unstable_ghosts"]),
            "num_remaining_ghosts": np.sum(result_acc["alive_ghosts"]),
            "common_ghosts": np.sum(
                np.logical_and(
                    result_acc["unstable_ghosts"], result_acc["alive_ghosts"]
                )
            ),
            "time_with_dropping": time_with_dropping,
            "time_without_dropping": time_without_dropping,
        }

        results.append(result)

    return results


def benchmark_v1(
    data_name: str,
    X: np.ndarray,
    precomputed_knn,
    hyperparameters,
    iterations=10,
):
    results = []

    for i in range(iterations):
        logging.info(f"Starting iteration {i + 1}/{iterations}.")
        result_acc_dropping = run_GhostUMAP(
            X,
            hyperparameters,
            bm_type="accuracy_dropping",
            precomputed_knn=precomputed_knn,
            distance=0.1,
        )
        save_embeddings(
            result_acc_dropping,
            hyperparameters,
            "accuracy_dropping",
            data_name,
            results_dir="results_v1",
        )

        result_acc_sh = run_GhostUMAP(
            X,
            hyperparameters,
            bm_type="accuracy_SH",
            precomputed_knn=precomputed_knn,
            distance=0.1,
        )
        save_embeddings(
            result_acc_sh,
            hyperparameters,
            "accuracy_dropping",
            data_name,
            results_dir="results_v1",
        )

        result_twd = run_GhostUMAP(
            X,
            hyperparameters,
            bm_type="time_with_dropping",
            precomputed_knn=precomputed_knn,
        )
        save_embeddings(
            result_twd,
            hyperparameters,
            "time_with_dropping",
            data_name,
            results_dir="results_v1",
        )

        result_twsh = run_GhostUMAP(
            X,
            hyperparameters,
            bm_type="time_with_SH",
            precomputed_knn=precomputed_knn,
        )
        save_embeddings(
            result_twd,
            hyperparameters,
            "time_with_SH",
            data_name,
            results_dir="results_v1",
        )

        result_time_orig = run_GhostUMAP(
            X,
            hyperparameters,
            bm_type="time_original_GU",
            precomputed_knn=precomputed_knn,
        )
        save_embeddings(
            result_time_orig,
            hyperparameters,
            "time_without_dropping",
            data_name,
            results_dir="results_v1",
        )

        true_drop = result_acc_dropping["unstable_ghosts"]
        pred_drop = drop_ghosts(
            result_acc_dropping["original_embedding"],
            result_acc_dropping["ghost_embedding"],
            result_acc_dropping["alive_ghosts"],
            sensitivity=hyperparameters["sensitivity"],
            distance=0.1,
        )

        f1_drop, precision_drop, recall_drop = measure_accuracy(true_drop, pred_drop)

        true_sh = result_acc_sh["unstable_ghosts"]
        pred_sh = drop_ghosts(
            result_acc_sh["original_embedding"],
            result_acc_sh["ghost_embedding"],
            result_acc_sh["alive_ghosts"],
            sensitivity=hyperparameters["sensitivity"],
            distance=0.1,
        )
        f1_sh, precision_sh, recall_sh = measure_accuracy(true_sh, pred_sh)

        print(f1_drop, precision_drop, recall_drop, f1_sh, precision_sh, recall_sh)
        time_with_dropping = result_twd["opt_time"]
        time_with_SH = result_twsh["opt_time"]
        time_GU = result_time_orig["opt_time"]

        result = {
            "data": data_name,
            **hyperparameters,
            "iter": i,
            "time_original_GU": time_GU,
            "time_with_dropping": time_with_dropping,
            "time_with_SH": time_with_SH,
            "f1_dropping": f1_drop,
            "precision_dropping": precision_drop,
            "recall_dropping": recall_drop,
            "f1_SH": f1_sh,
            "precision_SH": precision_sh,
            "recall_SH": recall_sh,
            "num_unstable_ghosts_dropping": np.sum(true_drop),
            "num_remaining_ghosts_dropping": np.sum(
                result_acc_dropping["alive_ghosts"]
            ),
            "num_unstable_ghosts_after_dropping": np.sum(pred_drop),
            "common_ghosts_dropping": np.sum(np.logical_and(true_drop, pred_drop)),
            "num_unstable_ghosts_SH": np.sum(true_sh),
            "num_remaining_ghosts_SH": np.sum(result_acc_sh["alive_ghosts"]),
            "num_unstable_ghosts_after_SH": np.sum(pred_sh),
            "common_ghosts_SH": np.sum(np.logical_and(true_sh, pred_sh)),
        }

        results.append(result)

    return results


def benchmark_v2(
    data_name: str,
    X: np.ndarray,
    precomputed_knn,
    hyperparameters,
    iterations=3,
):
    """
    Adding original UMAP and time comparison
    """
    results = []

    for i in range(iterations):
        logging.info(f"Starting iteration {i + 1}/{iterations}.")
        result_acc_dropping = run_GhostUMAP(
            X,
            hyperparameters,
            bm_type="accuracy_dropping",
            precomputed_knn=precomputed_knn,
            distance=0.1,
        )
        save_embeddings(
            result_acc_dropping,
            hyperparameters,
            "accuracy_dropping",
            data_name,
            results_dir="results_v2",
        )

        result_acc_sh = run_GhostUMAP(
            X,
            hyperparameters,
            bm_type="accuracy_SH",
            precomputed_knn=precomputed_knn,
            distance=0.1,
        )
        save_embeddings(
            result_acc_sh,
            hyperparameters,
            "accuracy_dropping",
            data_name,
            results_dir="results_v2",
        )

        result_twd = run_GhostUMAP(
            X,
            hyperparameters,
            bm_type="time_with_dropping",
            precomputed_knn=precomputed_knn,
        )
        save_embeddings(
            result_twd,
            hyperparameters,
            "time_with_dropping",
            data_name,
            results_dir="results_v2",
        )

        result_twsh = run_GhostUMAP(
            X,
            hyperparameters,
            bm_type="time_with_SH",
            precomputed_knn=precomputed_knn,
        )
        save_embeddings(
            result_twd,
            hyperparameters,
            "time_with_SH",
            data_name,
            results_dir="results_v2",
        )

        result_time_orig = run_GhostUMAP(
            X,
            hyperparameters,
            bm_type="time_original_GU",
            precomputed_knn=precomputed_knn,
        )
        save_embeddings(
            result_time_orig,
            hyperparameters,
            "time_without_dropping",
            data_name,
            results_dir="results_v2",
        )

        result_time_original_UMAP = run_GhostUMAP(
            X,
            hyperparameters,
            bm_type="original_UMAP",
            precomputed_knn=precomputed_knn,
        )
        save_embeddings(
            result_time_original_UMAP,
            hyperparameters,
            "original_UMAP",
            data_name,
            results_dir="results_v2",
        )

        true_drop = result_acc_dropping["unstable_ghosts"]
        pred_drop = drop_ghosts(
            result_acc_dropping["original_embedding"],
            result_acc_dropping["ghost_embedding"],
            result_acc_dropping["alive_ghosts"],
            sensitivity=hyperparameters["sensitivity"],
            distance=0.1,
        )

        f1_drop, precision_drop, recall_drop = measure_accuracy(true_drop, pred_drop)

        true_sh = result_acc_sh["unstable_ghosts"]
        pred_sh = drop_ghosts(
            result_acc_sh["original_embedding"],
            result_acc_sh["ghost_embedding"],
            result_acc_sh["alive_ghosts"],
            sensitivity=hyperparameters["sensitivity"],
            distance=0.1,
        )
        f1_sh, precision_sh, recall_sh = measure_accuracy(true_sh, pred_sh)

        print(f1_drop, precision_drop, recall_drop, f1_sh, precision_sh, recall_sh)
        time_with_dropping = result_twd["opt_time"]
        time_with_SH = result_twsh["opt_time"]
        time_GU = result_time_orig["opt_time"]
        time_UMAP = result_time_original_UMAP["opt_time"]

        result = {
            "data": data_name,
            **hyperparameters,
            "iter": i,
            "time_UMAP": time_UMAP,
            "time_original_GU": time_GU,
            "time_with_dropping": time_with_dropping,
            "time_with_SH": time_with_SH,
            "f1_dropping": f1_drop,
            "precision_dropping": precision_drop,
            "recall_dropping": recall_drop,
            "f1_SH": f1_sh,
            "precision_SH": precision_sh,
            "recall_SH": recall_sh,
            "num_unstable_ghosts_dropping": np.sum(true_drop),
            "num_remaining_ghosts_dropping": np.sum(
                result_acc_dropping["alive_ghosts"]
            ),
            "num_unstable_ghosts_after_dropping": np.sum(pred_drop),
            "common_ghosts_dropping": np.sum(np.logical_and(true_drop, pred_drop)),
            "num_unstable_ghosts_SH": np.sum(true_sh),
            "num_remaining_ghosts_SH": np.sum(result_acc_sh["alive_ghosts"]),
            "num_unstable_ghosts_after_SH": np.sum(pred_sh),
            "common_ghosts_SH": np.sum(np.logical_and(true_sh, pred_sh)),
        }

        results.append(result)

    return results


__all__ = ["benchmark_v0", "benchmark_v1", "benchmark_v2"]



from functools import wraps
import time


def measure_time(func_name: str):
    def decorator(func):
        @wraps(func)
        def wrapper(*args, **kwargs):
            start = time.time()
            res = func(*args, **kwargs)
            end = time.time()

            return res, end - start

        return wrapper

    return decorator



import os
from .model import TDataName, DataModel

from .mnistSeries import MnistSeriesLoader
from .celegans import CelegansLoader
from .uciml import UcimlLoader
from .cnae9 import Cnae9Loader
from .parishousing import ParisHousingLoader
from .textEmbeddings import TextEmbeddingLoader


class DataLoader:
    def __init__(self, data_name: TDataName = "mnist"):
        self.data_name = data_name
        self.loader = self._get_loader()

    def _get_loader(self):
        if self.data_name in ["mnist", "fmnist", "kmnist"]:
            return MnistSeriesLoader(self.data_name)
        elif self.data_name == "celegans":
            return CelegansLoader()
        elif self.data_name in [
            "ionosphere",
            "optical_recognition",
            "raisin",
            "htru2",
        ]:
            return UcimlLoader(self.data_name)
        elif self.data_name == "parishousing":
            return ParisHousingLoader()
        elif self.data_name == "cnae9":
            return Cnae9Loader()

        elif self.data_name in [
            "20ng",
            "ag_news",
            "amazon_polarity",
            "yelp_review",
        ]:
            return TextEmbeddingLoader(self.data_name)
        else:
            raise ValueError("Invalid type")

    def get_data(self) -> DataModel:
        return self.loader.get_data()



from dataclasses import dataclass, field, asdict
from typing import Literal, Optional


Tbenchmark = Literal[
    "None",
    "accuracy_dropping",
    "time_with_dropping",
    "accuracy_SH",
    "time_with_SH",
    "time_original_GU",
    "original_UMAP",
]


@dataclass(frozen=True)
class Config:
    r: float = field(default=0.1)
    sensitivity: float = field(default=1)
    ghost_gen: float = field(default=0.25)
    dropping: bool = field(default=True)
    init_dropping: float = field(default=0.5)
    smoothing_factor: float = field(default=0.9)
    bm_type: Tbenchmark = field(default="None")


_config: Optional[Config] = None


def get_config() -> Config:
    global _config
    if _config is None:
        raise ValueError("Config not set")
    return _config


def set_config(
    r,
    sensitivity,
    ghost_gen=0.2,
    dropping=True,
    init_dropping=0.4,
    smoothing_factor=0.9,
    bm_type: Tbenchmark = "None",
) -> None:
    global _config
    # if _config:
    #     raise ValueError("Config already exists")

    _config = Config(
        r=r,
        sensitivity=sensitivity,
        ghost_gen=ghost_gen,
        dropping=dropping,
        init_dropping=init_dropping,
        smoothing_factor=smoothing_factor,
        bm_type=bm_type,
    )


__all__ = ["get_config", "set_config"]



from dataclasses import asdict
import numpy as np
import numba

from umap.utils import tau_rand_int
from tqdm.auto import tqdm

from .time_checker import measure_time

from .utils import (
    _get_max_extent,
    compute_distances,
    drop_ghosts,
    get_distance,
    sample_ghosts,
)

from .configs import get_config
from .results import set_results


@numba.njit()
def clip(val):
    """Standard clamping of a value into a fixed range (in this case -4.0 to
    4.0)

    Parameters
    ----------
    val: float
        The value to be clamped.

    Returns
    -------
    The clamped value, now fixed to be in the range -4.0 to 4.0.
    """
    if val > 4.0:
        return 4.0
    elif val < -4.0:
        return -4.0
    else:
        return val


@numba.njit(
    "f4(f4[::1],f4[::1])",
    fastmath=True,
    cache=True,
    locals={
        "result": numba.types.float32,
        "diff": numba.types.float32,
        "dim": numba.types.intp,
        "i": numba.types.intp,
    },
)
def rdist(x, y):
    """Reduced Euclidean distance.

    Parameters
    ----------
    x: array of shape (embedding_dim,)
    y: array of shape (embedding_dim,)

    Returns
    -------
    The squared euclidean distance between x and y
    """
    result = 0.0
    dim = x.shape[0]
    for i in range(dim):
        diff = x[i] - y[i]
        result += diff * diff

    return result


def _optimize_real_layout_euclidean_single_epoch(
    head_embedding,
    tail_embedding,
    head,
    tail,
    n_vertices,
    epochs_per_sample,
    a,
    b,
    rng_state,
    gamma,
    dim,
    move_other,
    alpha,
    epochs_per_negative_sample,
    epoch_of_next_negative_sample,
    epoch_of_next_sample,
    n,
):
    for i in numba.prange(epochs_per_sample.shape[0]):
        if epoch_of_next_sample[i] <= n:
            j = head[i]
            k = tail[i]

            current = head_embedding[j]
            other = tail_embedding[k]

            dist_squared = rdist(current, other)

            if dist_squared > 0.0:
                grad_coeff = -2.0 * a * b * pow(dist_squared, b - 1.0)
                grad_coeff /= a * pow(dist_squared, b) + 1.0
            else:
                grad_coeff = 0.0

            for d in range(dim):
                grad_d = clip(grad_coeff * (current[d] - other[d]))

                current[d] += grad_d * alpha
                if move_other:
                    other[d] += -grad_d * alpha

            epoch_of_next_sample[i] += epochs_per_sample[i]

            n_neg_samples = int(
                (n - epoch_of_next_negative_sample[i]) / epochs_per_negative_sample[i]
            )

            for p in range(n_neg_samples):
                k = tau_rand_int(rng_state) % n_vertices

                other = tail_embedding[k]

                dist_squared = rdist(current, other)

                if dist_squared > 0.0:
                    grad_coeff = 2.0 * gamma * b
                    grad_coeff /= (0.001 + dist_squared) * (
                        a * pow(dist_squared, b) + 1
                    )
                elif j == k:
                    continue
                else:
                    grad_coeff = 0.0

                for d in range(dim):
                    if grad_coeff > 0.0:
                        grad_d = clip(grad_coeff * (current[d] - other[d]))
                    else:
                        grad_d = 0
                    current[d] += grad_d * alpha

            epoch_of_next_negative_sample[i] += (
                n_neg_samples * epochs_per_negative_sample[i]
            )


def _optimize_ghost_layout_euclidean_single_epoch(
    ghost_embeddings,
    n_ghosts,
    ghost_mask,
    head_embedding,
    tail_embedding,
    head,
    tail,
    n_vertices,
    epochs_per_sample,
    a,
    b,
    rng_state,
    gamma,
    dim,
    move_other,
    alpha,
    epochs_per_negative_sample,
    epoch_of_next_negative_sample,
    epoch_of_next_sample,
    n,
):
    """
    ghost_embedding: array of shape (n_vertices, n_ghosts_per_target, n_components)
        Embeddings of ghosts
    has_ghost: array of shape (n_vertices)
        Whether the vertex has a ghost
    num_ghosts_per_target: int
        number of ghosts per target
    """
    # optimize ghost embedding
    for i in numba.prange(epochs_per_sample.shape[0]):
        if epoch_of_next_sample[i] <= n:
            j = head[i]  # 1st node of the ith link
            k = tail[i]  # 2nd node of the ith link, link is symmetric
        else:
            continue

        if not ghost_mask[j]:
            continue

        current_orig = head_embedding[j]
        other_orig = tail_embedding[k]

        if ghost_mask[j]:
            for g in range(n_ghosts):
                current_ghost = ghost_embeddings[j][g]

                dist_squared = rdist(
                    current_ghost, other_orig
                )  # squared euclidean distance

                if dist_squared > 0.0:  # attractive force
                    grad_coeff = -2.0 * a * b * pow(dist_squared, b - 1.0)
                    grad_coeff /= a * pow(dist_squared, b) + 1.0
                else:
                    grad_coeff = 0.0

                for d in range(dim):
                    grad_d = clip(grad_coeff * (current_ghost[d] - other_orig[d]))
                    current_ghost[d] += grad_d * alpha

                n_neg_samples = int(
                    (n - epoch_of_next_negative_sample[i])
                    / epochs_per_negative_sample[i]
                )

                for p in range(n_neg_samples):
                    neg_k = (
                        tau_rand_int(rng_state) % n_vertices
                    )  # A fast (pseudo)-random number generator.

                    other_neg = tail_embedding[neg_k]

                    dist_squared = rdist(current_ghost, other_neg)

                    if dist_squared > 0.0:  # repulsive force
                        grad_coeff = 2.0 * gamma * b
                        grad_coeff /= (0.001 + dist_squared) * (
                            a * pow(dist_squared, b) + 1
                        )
                    elif j == neg_k:
                        continue
                    else:
                        grad_coeff = 0.0

                    for d in range(dim):
                        if grad_coeff > 0.0:
                            grad_d = clip(
                                grad_coeff * (current_ghost[d] - other_neg[d])
                            )
                        else:
                            grad_d = 0
                        current_ghost[d] += grad_d * alpha

        if ghost_mask[k] and move_other:
            for g in range(n_ghosts):
                other_ghost = ghost_embeddings[k][g]

                dist_squared = rdist(current_orig, other_ghost)

                if dist_squared > 0.0:  # attractive force
                    grad_coeff = -2.0 * a * b * pow(dist_squared, b - 1.0)
                    grad_coeff /= a * pow(dist_squared, b) + 1.0
                else:
                    grad_coeff = 0.0

                for d in range(dim):
                    grad_d = clip(grad_coeff * (current_orig[d] - other_ghost[d]))
                    other_ghost[d] += -grad_d * alpha


def optimize_layout_euclidean(
    n_ghosts,
    original_embedding,
    head,
    tail,
    n_epochs,
    n_vertices,
    epochs_per_sample,
    a,
    b,
    rng_state,
    gamma=1.0,
    initial_alpha=1.0,
    negative_sample_rate=5.0,
    parallel=False,
    verbose=False,
    tqdm_kwds=None,
    move_other=False,
):
    """Improve an embedding using stochastic gradient descent to minimize the
    fuzzy set cross entropy between the 1-skeletons of the high dimensional
    and low dimensional fuzzy simplicial sets. In practice this is done by
    sampling edges based on their membership strength (with the (1-p) terms
    coming from negative sampling similar to word2vec).
    Parameters
    ----------
    n_embeddings: int
        The number of embeddings to optimize
    n_ghosts: int
        The number of ghosts per target
    original_embeddings: array of shape (n_embeddings, n_samples, n_components)
        The initial embeddings to be improved by SGD.
    head_embedding: array of shape (n_samples, n_components)
        The initial embedding to be improved by SGD.
    tail_embedding: array of shape (source_samples, n_components)
        The reference embedding of embedded points. If not embedding new
        previously unseen points with respect to an existing embedding this
        is simply the head_embedding (again); otherwise it provides the
        existing embedding to embed with respect to.
    head: array of shape (n_1_simplices)
        The indices of the heads of 1-simplices with non-zero membership.
    tail: array of shape (n_1_simplices)
        The indices of the tails of 1-simplices with non-zero membership.
    n_epochs: int, or list of int
        The number of training epochs to use in optimization, or a list of
        epochs at which to save the embedding. In case of a list, the optimization
        will use the maximum number of epochs in the list, and will return a list
        of embedding in the order of increasing epoch, regardless of the order in
        the epoch list.
    n_vertices: int
        The number of vertices (0-simplices) in the dataset.
    epochs_per_sample: array of shape (n_1_simplices)
        A float value of the number of epochs per 1-simplex. 1-simplices with
        weaker membership strength will have more epochs between being sampled.
    a: float
        Parameter of differentiable approximation of right adjoint functor
    b: float
        Parameter of differentiable approximation of right adjoint functor
    rng_state: array of int64, shape (3,)
        The internal state of the rng
    gamma: float (optional, default 1.0)
        Weight to apply to negative samples.
    initial_alpha: float (optional, default 1.0)
        Initial learning rate for the SGD.
    negative_sample_rate: int (optional, default 5)
        Number of negative samples to use per positive sample.
    parallel: bool (optional, default False)
        Whether to run the computation using numba parallel.
        Running in parallel is non-deterministic, and is not used
        if a random seed has been set, to ensure reproducibility.
    verbose: bool (optional, default False)
        Whether to report information on the current progress of the algorithm.
    densmap: bool (optional, default False)
        Whether to use the density-augmented densMAP objective
    densmap_kwds: dict (optional, default None)
        Auxiliary data for densMAP
    tqdm_kwds: dict (optional, default None)
        Keyword arguments for tqdm progress bar.
    move_other: bool (optional, default False)
        Whether to adjust tail_embedding alongside head_embedding
    Returns
    -------
    embedding: array of shape (n_samples, n_components)
        The optimized embedding.
    """

    config = get_config()
    (
        radii,
        sensitivity,
        ghost_gen,
        dropping,
        init_dropping,
        mov_avg_weight,
        bm_type,
    ) = asdict(config).values()

    dim = original_embedding.shape[1]
    alpha = initial_alpha

    epochs_per_negative_sample = epochs_per_sample / negative_sample_rate
    epoch_of_next_negative_sample = epochs_per_negative_sample.copy()
    epoch_of_next_sample = epochs_per_sample.copy()

    optimize_real_fn = numba.njit(
        _optimize_real_layout_euclidean_single_epoch,
        fastmath=True,
        parallel=parallel,
    )

    optimize_ghost_fn = numba.njit(
        _optimize_ghost_layout_euclidean_single_epoch,
        fastmath=True,
        parallel=parallel,
    )

    if tqdm_kwds is None:
        tqdm_kwds = {}

    epochs_list = None

    if isinstance(n_epochs, list):
        epochs_list = n_epochs
        n_epochs = max(epochs_list)

    if "disable" not in tqdm_kwds:
        tqdm_kwds["disable"] = not verbose

    ghost_embeddings = None

    # for dropping ghosts
    ghost_mask = np.ones(n_vertices, dtype=np.bool_)
    distance_list = []
    thresholds = []

    init_radii = None

    for n in tqdm(range(n_epochs), **tqdm_kwds):
        if ghost_embeddings is None and n >= int(n_epochs * ghost_gen):
            if verbose:
                print(f"Generating ghosts at epoch {n}")
            ghost_embeddings = sample_ghosts(original_embedding, n_ghosts, r=radii)
            init_radii = compute_distances(
                original_embedding, ghost_embeddings
            ) / _get_max_extent(original_embedding)

        if ghost_embeddings is not None:
            optimize_ghost_fn(
                ghost_embeddings,
                n_ghosts,
                ghost_mask,
                original_embedding.astype(np.float32),
                original_embedding.astype(np.float32),
                head,
                tail,
                n_vertices,
                epochs_per_sample,
                a,
                b,
                rng_state,
                gamma,
                dim,
                move_other,
                alpha,
                epochs_per_negative_sample,
                epoch_of_next_negative_sample,
                epoch_of_next_sample,
                n,
            )

        optimize_real_fn(
            original_embedding,
            original_embedding,
            head,
            tail,
            n_vertices,
            epochs_per_sample,
            a,
            b,
            rng_state,
            gamma,
            dim,
            move_other,
            alpha,
            epochs_per_negative_sample,
            epoch_of_next_negative_sample,
            epoch_of_next_sample,
            n,
        )

        alpha = initial_alpha * (1.0 - (float(n) / float(n_epochs)))

        if verbose and n % int(n_epochs / 10) == 0:
            print("\tcompleted ", n, " / ", n_epochs, "epochs")

        if ghost_embeddings is not None and dropping:
            _update_distances(
                distance_list,
                thresholds,
                original_embedding,
                ghost_embeddings,
                ghost_mask,
                sensitivity,
                mov_avg_weight,
            )

            _drop_ghosts_if_needed(
                n,
                n_epochs,
                init_dropping,
                ghost_mask,
                distance_list,
                thresholds,
            )

    set_results(
        distance_list=np.array(distance_list),
        threshold_list=np.array(thresholds),
        init_radii=init_radii,
    )

    return (original_embedding, ghost_embeddings, ghost_mask)


def _update_distances(
    distance_list: list,
    thresholds: list,
    original_embedding,
    ghost_embeddings,
    ghost_mask,
    sensitivity,
    mov_avg_weight,
):
    if not distance_list:
        distance_list.append(
            get_distance(
                original_embedding,
                ghost_embeddings,
                ghost_mask,
                sensitivity=sensitivity,
            )
        )

        thresholds.append(np.mean(distance_list[-1]))

    else:
        distances = np.copy(distance_list[-1])
        distances[ghost_mask] = get_distance(
            original_embedding,
            ghost_embeddings,
            ghost_mask,
            sensitivity=sensitivity,
        )

        mov_avg_dist = (
            np.copy(distance_list[-1]) * mov_avg_weight
            + (1 - mov_avg_weight) * distances
        )

        distance_list.append(mov_avg_dist)
        thresholds.append(np.mean(mov_avg_dist))


def _drop_ghosts_if_needed(
    epoch, n_epochs, init_dropping, ghost_mask, distance_list, thresholds
):
    if epoch >= int(n_epochs * init_dropping):
        active_indices = np.where(ghost_mask)[0]
        ghost_to_drop = distance_list[-1][ghost_mask] < thresholds[-1]

        ghost_mask[active_indices[ghost_to_drop]] = False
    return ghost_mask



from datetime import datetime
import os
from typing import Dict, List

import numpy as np
import pandas as pd

from .hyperparameters import make_dir_name
from .models import THyperparameter, TResult, Tbenchmark


def save_embeddings(
    result: TResult,
    hyperparameters: THyperparameter,
    bm_type: Tbenchmark,
    data_name: str,
    results_dir: str = "results",
):
    if not os.path.isdir(results_dir):
        os.makedirs(results_dir)

    base_path = os.path.join(results_dir, data_name)

    if not os.path.isdir(base_path):
        os.makedirs(base_path)

    dir_name = make_dir_name(hyperparameters)
    hpram_path = os.path.join(base_path, dir_name)

    if not os.path.isdir(hpram_path):
        os.makedirs(hpram_path)

    eval_path = os.path.join(hpram_path, bm_type)

    if not os.path.isdir(eval_path):
        os.makedirs(eval_path)

    cnt = len(os.listdir(eval_path)) + 1

    result = {key: val for key, val in result.items() if key != "opt_time"}

    np.savez_compressed(f"{eval_path}/{cnt}.npz", **result)

    return


def save_results(
    data_name: str,
    results: List[Dict],
    results_dir: str = "results",
):
    df = pd.DataFrame(results)

    if not os.path.isdir(results_dir):
        os.makedirs(results_dir)

    if not os.path.isdir(f"{results_dir}/{data_name}"):
        os.makedirs(f"{results_dir}/{data_name}")

    if not os.path.exists(f"{results_dir}/result.csv"):
        df.to_csv(f"{results_dir}/result.csv", index=False)
    else:
        overall_df_old = pd.read_csv(f"{results_dir}/result.csv")
        overall_df = pd.concat([overall_df_old, df], ignore_index=True)
        overall_df.to_csv(f"{results_dir}/result.csv", index=False)

    if not os.path.exists(f"{results_dir}/{data_name}/result.csv"):
        df.to_csv(f"{results_dir}/{data_name}/result.csv", index=False)

    else:
        individual_df_old = pd.read_csv(f"{results_dir}/{data_name}/result.csv")
        individual_df = pd.concat([individual_df_old, df], ignore_index=True)
        individual_df.to_csv(f"{results_dir}/{data_name}/result.csv", index=False)



from .EmbeddingSet import EmbeddingSet
from .OriginalEmbedding import OriginalPointModel, OriginalEmbedding
from .GhostEmbedding import GhostPointModel, GhostEmbedding


__all__ = [
    "EmbeddingSet",
    "OriginalPointModel",
    "GhostPointModel",
    "OriginalEmbedding",
    "GhostEmbedding",
]



import json
import os
from abc import ABC, abstractmethod

import numpy as np
from umap import UMAP

from .model import DataModel


class BaseLoader(ABC):
    def __init__(self):
        self.base_path = None
        self._data = None
        self._label = None
        self._legend = None
        self._precomputed_knn = None

    @abstractmethod
    def load_raw_data(self):
        pass

    def drop_cache(self, paths):
        for path in paths:
            if os.path.exists(path):
                os.remove(path)

    def load_data(self):
        paths = {
            "data": os.path.join(self.base_path, "data.npy"),
            "label": os.path.join(self.base_path, "label.npy"),
            "legend": os.path.join(self.base_path, "legend.json"),
            "knn_indices": os.path.join(self.base_path, "knn_indices.npy"),
            "knn_dists": os.path.join(self.base_path, "knn_dists.npy"),
        }

        if not all(os.path.exists(path) for path in paths.values()):
            self.load_raw_data()
            return

        self._data = np.load(paths["data"], mmap_mode=None)
        self._label = np.load(paths["label"], mmap_mode=None)
        self._precomputed_knn = (
            (
                np.load(paths["knn_indices"], mmap_mode=None),
                np.load(paths["knn_dists"], mmap_mode=None),
            )
            if os.path.exists(paths["knn_indices"])
            and os.path.exists(paths["knn_dists"])
            else (None, None)
        )

        self._legend = (
            json.load(open(paths["legend"])).get("legend")
            if os.path.exists(paths["legend"])
            else None
        )

    def get_data(self) -> DataModel:
        result = {
            "data": self._data,
            "label": self._label,
            "legend": self._legend,
            "precomputed_knn": self._precomputed_knn,
        }

        return result

    def save_data(self, path):
        save_if_not_exists(path, self._data)
        save_if_not_exists(path, self._label)

        if self._precomputed_knn[0] is not None:
            save_if_not_exists(
                os.path.join(path, "knn_indices.npy"), self._precomputed_knn[0]
            )
            save_if_not_exists(
                os.path.join(path, "knn_dists.npy"), self._precomputed_knn[1]
            )

        legend_path = os.path.join(path, "legend.json")
        if not os.path.exists(legend_path):
            with open(os.path.join(path, "legend.json"), "w") as f:
                json.dump({"legend": self._legend}, f)

    def scale_data(self, data: np.ndarray):
        from sklearn.preprocessing import StandardScaler

        scaler = StandardScaler()
        return scaler.fit_transform(data)

    def get_precomputed_knn(self, X, n_neighbors: int = 15):
        if X.shape[0] < 4096:
            self._precomputed_knn = (None, None)
            return self._precomputed_knn

        reducer = UMAP(n_neighbors=n_neighbors)
        _ = reducer.fit_transform(X)
        self._precomputed_knn = (reducer._knn_indices, reducer._knn_dists)

        return self._precomputed_knn


def save_if_not_exists(file_path, data):
    if not os.path.exists(file_path):
        np.save(file_path, data)
